import logging
from typing import TYPE_CHECKING, Callable, Dict, List, NamedTuple, Optional, Tuple, Union

from eth_utils.address import to_checksum_address
from typing_extensions import Literal

from rotkehlchen.accounting.structures import Balance
from rotkehlchen.assets.asset import Asset, EthereumToken
from rotkehlchen.chain.ethereum.contracts import EthereumContract
from rotkehlchen.chain.ethereum.defi import handle_defi_price_query
from rotkehlchen.chain.ethereum.utils import token_normalized_value_decimals
from rotkehlchen.constants.ethereum import ZERION_ABI
from rotkehlchen.constants.misc import ZERO
from rotkehlchen.errors import UnknownAsset, UnsupportedAsset
from rotkehlchen.fval import FVal
from rotkehlchen.inquirer import Inquirer, get_underlying_asset_price
from rotkehlchen.serialization.deserialize import deserialize_ethereum_address
from rotkehlchen.typing import ChecksumEthAddress, Price
from rotkehlchen.user_messages import MessagesAggregator

if TYPE_CHECKING:
    from rotkehlchen.chain.ethereum.manager import EthereumManager

log = logging.getLogger(__name__)


def _is_token_non_standard(symbol: str, address: ChecksumEthAddress) -> bool:
    """ignore some assets we do not yet support and don't want to spam warnings with"""
    if symbol in ('UNI-V2', 'swUSD', 'crCRV'):
        return True

    if address in ('0xCb2286d9471cc185281c4f763d34A962ED212962',):  # Sushi LP token
        return True

    return False


class DefiProtocol(NamedTuple):
    name: str
    description: str
    url: str
    version: int

    def serialize(self) -> Dict[str, str]:
        return {'name': self.name}


class DefiBalance(NamedTuple):
    token_address: ChecksumEthAddress
    token_name: str
    token_symbol: str
    balance: Balance


class DefiProtocolBalances(NamedTuple):
    protocol: DefiProtocol
    balance_type: Literal['Asset', 'Debt']
    base_balance: DefiBalance
    underlying_balances: List[DefiBalance]


# Type of the argument given to functions that need the defi balances
GIVEN_DEFI_BALANCES = Union[
    Dict[ChecksumEthAddress, List[DefiProtocolBalances]],
    Callable[[], Dict[ChecksumEthAddress, List[DefiProtocolBalances]]],
]


def _handle_pooltogether(normalized_balance: FVal, token_name: str) -> Optional[DefiBalance]:
    """Special handling for pooltogether

    https://github.com/rotki/rotki/issues/1429
    """
    if 'DAI' in token_name:
        dai_price = Inquirer.find_usd_price(EthereumToken('DAI'))
        return DefiBalance(
            token_address=to_checksum_address('0x49d716DFe60b37379010A75329ae09428f17118d'),
            token_name='Pool Together DAI token',
            token_symbol='plDAI',
            balance=Balance(
                amount=normalized_balance,
                usd_value=normalized_balance * dai_price,
            ),
        )
    elif 'USDC' in token_name:
        usdc_price = Inquirer.find_usd_price(EthereumToken('USDC'))
        return DefiBalance(
            token_address=to_checksum_address('0xBD87447F48ad729C5c4b8bcb503e1395F62e8B98'),
            token_name='Pool Together USDC token',
            token_symbol='plUSDC',
            balance=Balance(
                amount=normalized_balance,
                usd_value=normalized_balance * usdc_price,
            ),
        )

    return None


# supported zerion adapter address
ZERION_ADAPTER_ADDRESS = deserialize_ethereum_address('0x06FE76B2f432fdfEcAEf1a7d4f6C3d41B5861672')


class Zerion():
    """Adapter for the Zerion DeFi SDK https://github.com/zeriontech/defi-sdk"""

    def __init__(
            self,
            ethereum_manager: 'EthereumManager',
            msg_aggregator: MessagesAggregator,
    ) -> None:
        self.ethereum = ethereum_manager
        self.msg_aggregator = msg_aggregator
        self.contract = EthereumContract(
            address=ZERION_ADAPTER_ADDRESS,
            abi=ZERION_ABI,
            deployed_block=1586199170,
        )

    def all_balances_for_account(self, account: ChecksumEthAddress) -> List[DefiProtocolBalances]:
        """Calls the contract's getBalances() to get all protocol balances for account

        https://docs.zerion.io/smart-contracts/adapterregistry-v3#getbalances
        """
        result = self.contract.call(
            ethereum=self.ethereum,
            method_name='getBalances',
            arguments=[account],
        )
        protocol_balances = []
        for entry in result:
            protocol = DefiProtocol(
                name=entry[0][0],
                description=entry[0][1],
                url=entry[0][2],
                version=entry[0][4],
            )
            for adapter_balance in entry[1]:
                balance_type = adapter_balance[0][1]  # can be either 'Asset' or 'Debt'
                for balances in adapter_balance[1]:
                    underlying_balances = []
                    base_balance = self._get_single_balance(protocol.name, balances[0])
                    for balance in balances[1]:
                        defi_balance = self._get_single_balance(protocol.name, balance)
                        underlying_balances.append(defi_balance)

                    if base_balance.balance.usd_value == ZERO:
                        # This can happen. We can't find a price for some assets
                        # such as combined pool assets. But we can instead use
                        # the sum of the usd_value of the underlying_balances
                        usd_sum = sum(x.balance.usd_value for x in underlying_balances)
                        base_balance.balance.usd_value = usd_sum  # type: ignore

                    protocol_balances.append(DefiProtocolBalances(
                        protocol=protocol,
                        balance_type=balance_type,
                        base_balance=base_balance,
                        underlying_balances=underlying_balances,
                    ))

        return protocol_balances

    def _get_single_balance(
            self,
            protocol_name: str,
            entry: Tuple[Tuple[str, str, str, int], int],
    ) -> DefiBalance:
        metadata = entry[0]
        balance_value = entry[1]
        decimals = metadata[3]
        normalized_value = token_normalized_value_decimals(balance_value, decimals)
        token_symbol = metadata[2]
        token_address = to_checksum_address(metadata[0])
        token_name = metadata[1]

        special_handling = self.handle_protocols(
            protocol_name=protocol_name,
            token_symbol=token_symbol,
            normalized_balance=normalized_value,
            token_address=token_address,
            token_name=token_name,
        )
        if special_handling:
            return special_handling

        try:
            asset = Asset(token_symbol)
            usd_price = Inquirer().find_usd_price(asset)
        except (UnknownAsset, UnsupportedAsset):
            if not _is_token_non_standard(token_symbol, token_address):
                self.msg_aggregator.add_warning(
                    f'Unsupported asset {token_symbol} with address '
                    f'{token_address} encountered during DeFi protocol queries',
                )
            usd_price = Price(ZERO)

        usd_value = normalized_value * usd_price
        defi_balance = DefiBalance(
            token_address=token_address,
            token_name=token_name,
            token_symbol=token_symbol,
            balance=Balance(amount=normalized_value, usd_value=usd_value),
        )
        return defi_balance

    def handle_protocols(
            self,
            protocol_name: str,
            token_symbol: str,
            normalized_balance: FVal,
            token_address: str,
            token_name: str,
    ) -> Optional[DefiBalance]:
        """Special handling for price for token/protocols which are easier to do onchain
        or need some kind of special treatment.
        """
        if protocol_name == 'PoolTogether':
            result = _handle_pooltogether(normalized_balance, token_name)
            if result is not None:
                return result

        underlying_asset_price = get_underlying_asset_price(token_symbol)
        usd_price = handle_defi_price_query(self.ethereum, token_symbol, underlying_asset_price)
        if usd_price is None:
            return None

        return DefiBalance(
            token_address=to_checksum_address(token_address),
            token_name=token_name,
            token_symbol=token_symbol,
            balance=Balance(amount=normalized_balance, usd_value=normalized_balance * usd_price),
        )



import csv
from itertools import count
from pathlib import Path
from typing import Any, Dict, List

from rotkehlchen.assets.asset import Asset
from rotkehlchen.assets.asset import Asset, EthereumToken
from rotkehlchen.constants.misc import ZERO
from rotkehlchen.db.dbhandler import DBHandler
from rotkehlchen.errors import DeserializationError, UnknownAsset
from rotkehlchen.exchanges.data_structures import AssetMovement, AssetMovementCategory, Trade
from rotkehlchen.serialization.deserialize import (
    deserialize_asset_amount,
    deserialize_asset_amount_force_positive,
    deserialize_asset_movement_category,
    deserialize_fee,
    deserialize_timestamp_from_date,
)
from rotkehlchen.typing import AssetAmount, Fee, Location, Price, TradePair, TradeType


def remap_header(fieldnames: List[str]) -> List[str]:
    cur_count = count(1)
    mapping = {1: 'Buy', 2: 'Sell', 3: 'Fee'}
    return [f'Cur.{mapping[next(cur_count)]}' if f.startswith('Cur.') else f for f in fieldnames]


class UnsupportedCointrackingEntry(Exception):
    """Thrown for Cointracking CSV export entries we can't support to import"""


class UnsupportedCryptocomEntry(Exception):
    """Thrown for Cryptocom CSV export entries we can't support to import"""


def exchange_row_to_location(entry: str) -> Location:
    """Takes the exchange row entry of Cointracking exported trades list and returns a location"""
    if entry == 'no exchange':
        return Location.EXTERNAL
    elif entry == 'Kraken':
        return Location.KRAKEN
    elif entry == 'Poloniex':
        return Location.POLONIEX
    elif entry == 'Bittrex':
        return Location.BITTREX
    elif entry == 'Binance':
        return Location.BINANCE
    elif entry == 'Bitmex':
        return Location.BITMEX
    elif entry == 'Coinbase':
        return Location.COINBASE
    # TODO: Check if this is the correct string for CoinbasePro from cointracking
    elif entry == 'CoinbasePro':
        return Location.COINBASEPRO
    # TODO: Check if this is the correct string for Gemini from cointracking
    elif entry == 'Gemini':
        return Location.GEMINI
    elif entry == 'ETH Transaction':
        raise UnsupportedCointrackingEntry(
            'Not importing ETH Transactions from Cointracking. Cointracking does not '
            'export enough data for them. Simply enter your ethereum accounts and all '
            'your transactions will be auto imported directly from the chain',
        )
    elif entry == 'BTC Transaction':
        raise UnsupportedCointrackingEntry(
            'Not importing BTC Transactions from Cointracking. Cointracking does not '
            'export enough data for them. Simply enter your BTC accounts and all '
            'your transactions will be auto imported directly from the chain',
        )

    raise UnsupportedCointrackingEntry(
        f'Unknown Exchange "{entry}" encountered during a cointracking import. Ignoring it',
    )


class DataImporter():

    def __init__(self, db: DBHandler) -> None:
        self.db = db

    def _consume_cointracking_entry(self, csv_row: Dict[str, Any]) -> None:
        """Consumes a cointracking entry row from the CSV and adds it into the database
        Can raise:
            - DeserializationError if something is wrong with the format of the expected values
            - UnsupportedCointrackingEntry if importing of this entry is not supported.
            - IndexError if the CSV file is corrupt
            - KeyError if the an expected CSV key is missing
            - UnknownAsset if one of the assets founds in the entry are not supported
        """
        row_type = csv_row['Type']
        timestamp = deserialize_timestamp_from_date(
            date=csv_row['Date'],
            formatstr='%d.%m.%Y %H:%M:%S',
            location='cointracking.info',
        )
        notes = csv_row['Comment']
        location = exchange_row_to_location(csv_row['Exchange'])

        fee = Fee(ZERO)
        fee_currency = Asset('USD')  # whatever (used only if there is no fee)
        if csv_row['Fee'] != '':
            fee = deserialize_fee(csv_row['Fee'])
            fee_currency = Asset(csv_row['Cur.Fee'])

        if row_type in ('Gift/Tip', 'Trade', 'Income'):
            base_asset = Asset(csv_row['Cur.Buy'])
            quote_asset = None if csv_row['Cur.Sell'] == '' else Asset(csv_row['Cur.Sell'])
            if quote_asset is None and row_type not in ('Gift/Tip', 'Income'):
                raise DeserializationError('Got a trade entry with an empty quote asset')

            if quote_asset is None:
                # Really makes no difference as this is just a gift and the amount is zero
                quote_asset = Asset('USD')
            pair = TradePair(f'{base_asset.identifier}_{quote_asset.identifier}')
            base_amount_bought = deserialize_asset_amount(csv_row['Buy'])
            if csv_row['Sell'] != '-':
                quote_amount_sold = deserialize_asset_amount(csv_row['Sell'])
            else:
                quote_amount_sold = AssetAmount(ZERO)
            rate = Price(quote_amount_sold / base_amount_bought)

            trade = Trade(
                timestamp=timestamp,
                location=location,
                pair=pair,
                trade_type=TradeType.BUY,  # It's always a buy during cointracking import
                amount=base_amount_bought,
                rate=rate,
                fee=fee,
                fee_currency=fee_currency,
                link='',
                notes=notes,
            )
            self.db.add_trades([trade])
        elif row_type == 'Deposit' or row_type == 'Withdrawal':
            category = deserialize_asset_movement_category(row_type.lower())
            if category == AssetMovementCategory.DEPOSIT:
                amount = deserialize_asset_amount(csv_row['Buy'])
                asset = Asset(csv_row['Cur.Buy'])
            else:
                amount = deserialize_asset_amount_force_positive(csv_row['Sell'])
                asset = Asset(csv_row['Cur.Sell'])

            asset_movement = AssetMovement(
                location=location,
                category=category,
                address=None,
                transaction_id=None,
                timestamp=timestamp,
                asset=asset,
                amount=amount,
                fee=fee,
                fee_asset=fee_currency,
                link='',
            )
            self.db.add_asset_movements([asset_movement])
        else:
            raise UnsupportedCointrackingEntry(
                f'Unknown entrype type "{row_type}" encountered during cointracking '
                f'data import. Ignoring entry',
            )

    def import_cointracking_csv(self, filepath: Path) -> None:
        with open(filepath, 'r', encoding='utf-8-sig') as csvfile:
            data = csv.reader(csvfile, delimiter=',', quotechar='"')
            header = remap_header(next(data))
            for row in data:
                try:
                    self._consume_cointracking_entry(dict(zip(header, row)))
                except UnknownAsset as e:
                    self.db.msg_aggregator.add_warning(
                        f'During cointracking CSV import found action with unknown '
                        f'asset {e.asset_name}. Ignoring entry',
                    )
                    continue
                except IndexError:
                    self.db.msg_aggregator.add_warning(
                        'During cointracking CSV import found entry with '
                        'unexpected number of columns',
                    )
                    continue
                except DeserializationError as e:
                    self.db.msg_aggregator.add_warning(
                        f'Error during cointracking CSV import deserialization. '
                        f'Error was {str(e)}. Ignoring entry',
                    )
                    continue
                except UnsupportedCointrackingEntry as e:
                    self.db.msg_aggregator.add_warning(str(e))
                    continue

        return None

    def _consume_cryptocom_entry(self, csv_row: Dict[str, Any]) -> None:
        """Consumes a cryptocom entry row from the CSV and adds it into the database
        Can raise:
            - DeserializationError if something is wrong with the format of the expected values
            - UnsupportedCryptocomEntry if importing of this entry is not supported.
            - KeyError if the an expected CSV key is missing
            - UnknownAsset if one of the assets founds in the entry are not supported
        """
        row_type = csv_row['Transaction Kind']
        timestamp = deserialize_timestamp_from_date(
            date=csv_row['Timestamp (UTC)'],
            formatstr='%Y-%m-%d %H:%M:%S',
            location='crypto.com',
        )
        description = csv_row['Transaction Description']
        notes = f'{description}\nSource: crypto.com (CSV import)'

        # No fees info until (Nov 2020) on crypto.com
        # fees are not displayed in the export data
        fee = Fee(ZERO)
        fee_currency = Asset('USD')  # whatever (used only if there is no fee)

        if row_type in (
            'crypto_purchase',
            'crypto_exchange',
            'referral_gift',
            'referral_bonus',
            'crypto_earn_interest_paid',
            'referral_card_cashback',
            'card_cashback_reverted',
            'reimbursement',
        ):
            # variable mapping to raw data
            currency = csv_row['Currency']
            to_currency = csv_row['To Currency']
            native_currency = csv_row['Native Currency']
            amount = csv_row['Amount']
            to_amount = csv_row['To Amount']
            native_amount = csv_row['Native Amount']

            trade_type = TradeType.BUY if to_currency != native_currency else TradeType.SELL

            if row_type == 'crypto_exchange':
                # trades crypto to crypto
                base_asset = Asset(to_currency)
                quote_asset = Asset(currency)
                if quote_asset is None:
                    raise DeserializationError('Got a trade entry with an empty quote asset')
                base_amount_bought = deserialize_asset_amount(to_amount)
                quote_amount_sold = deserialize_asset_amount(amount)
            else:
                base_asset = Asset(currency)
                quote_asset = Asset(native_currency)
                base_amount_bought = deserialize_asset_amount(amount)
                quote_amount_sold = deserialize_asset_amount(native_amount)

            rate = Price(abs(quote_amount_sold / base_amount_bought))
            pair = TradePair(f'{base_asset.identifier}_{quote_asset.identifier}')
            trade = Trade(
                timestamp=timestamp,
                location=Location.CRYPTOCOM,
                pair=pair,
                trade_type=trade_type,
                amount=base_amount_bought,
                rate=rate,
                fee=fee,
                fee_currency=fee_currency,
                link='',
                notes=notes,
            )
            self.db.add_trades([trade])

        elif row_type == 'crypto_withdrawal' or row_type == 'crypto_deposit':
            if row_type == 'crypto_withdrawal':
                category = AssetMovementCategory.WITHDRAWAL
                amount = deserialize_asset_amount_force_positive(csv_row['Amount'])
            else:
                category = AssetMovementCategory.DEPOSIT
                amount = deserialize_asset_amount(csv_row['Amount'])

            asset = Asset(csv_row['Currency'])
            asset_movement = AssetMovement(
                location=Location.CRYPTOCOM,
                category=category,
                address=None,
                transaction_id=None,
                timestamp=timestamp,
                asset=asset,
                amount=amount,
                fee=fee,
                fee_asset=asset,
                link='',
            )
            self.db.add_asset_movements([asset_movement])

        elif row_type in (
            'crypto_earn_program_created',
            'lockup_lock',
            'lockup_unlock',
            'dynamic_coin_swap_bonus_exchange_deposit',
            'crypto_wallet_swap_debited',
            'crypto_wallet_swap_credited',
            'lockup_swap_debited',
            'lockup_swap_credited',
            'lockup_swap_rebate',
            'dynamic_coin_swap_debited',
            'dynamic_coin_swap_credited',
            'dynamic_coin_swap_bonus_exchange_deposit',
            # we don't handle cryto.com exchange yet
            'crypto_to_exchange_transfer',
            'exchange_to_crypto_transfer',
        ):
            # those types are ignored because it doesn't affect the wallet balance
            # or are not handled here
            return
        else:
            raise UnsupportedCryptocomEntry(
                f'Unknown entrype type "{row_type}" encountered during '
                f'cryptocom data import. Ignoring entry',
            )

    def _import_cryptocom_swap(self, data: Any) -> None:
        """Look for swapping events and handle them as trades.

        Notice: Crypto.com csv export gathers all swapping entries (`lockup_swap_*`,
        `crypto_wallet_swap_*`, ...) into one entry named `dynamic_coin_swap_*`.
        This method looks for `dynamic_coin_swap_debited` and `dynamic_coin_swap_credited`
        entries using the same timestamp to handle them as one trade.
        """
        swapping_rows: Dict[Any, Dict[str, Any]] = {}
        debited_row = None
        credited_row = None
        for row in data:
            if row['Transaction Kind'] == 'dynamic_coin_swap_debited':
                timestamp = deserialize_timestamp_from_date(
                    date=row['Timestamp (UTC)'],
                    formatstr='%Y-%m-%d %H:%M:%S',
                    location='crypto.com',
                )
                if timestamp not in swapping_rows:
                    swapping_rows[timestamp] = {}
                swapping_rows[timestamp]['debited'] = row
            elif row['Transaction Kind'] == 'dynamic_coin_swap_credited':
                timestamp = deserialize_timestamp_from_date(
                    date=row['Timestamp (UTC)'],
                    formatstr='%Y-%m-%d %H:%M:%S',
                    location='crypto.com',
                )
                if timestamp not in swapping_rows:
                    swapping_rows[timestamp] = {}
                swapping_rows[timestamp]['credited'] = row

        for timestamp in swapping_rows:
            credited_row = swapping_rows[timestamp]['credited']
            debited_row = swapping_rows[timestamp]['debited']
            if credited_row is not None and debited_row is not None:
                notes = 'Coin Swap\nSource: crypto.com (CSV import)'
                # No fees here since it's coin swapping
                fee = Fee(ZERO)
                fee_currency = Asset('USD')

                base_asset = Asset(credited_row['Currency'])
                quote_asset = Asset(debited_row['Currency'])
                pair = TradePair(f'{base_asset.identifier}_{quote_asset.identifier}')
                base_amount_bought = deserialize_asset_amount(credited_row['Amount'])
                quote_amount_sold = deserialize_asset_amount(debited_row['Amount'])
                rate = Price(abs(base_amount_bought / quote_amount_sold))

                trade = Trade(
                    timestamp=timestamp,
                    location=Location.CRYPTOCOM,
                    pair=pair,
                    trade_type=TradeType.BUY,
                    amount=base_amount_bought,
                    rate=rate,
                    fee=fee,
                    fee_currency=fee_currency,
                    link='',
                    notes=notes,
                )
                self.db.add_trades([trade])

    def import_cryptocom_csv(self, filepath: Path) -> None:
        with open(filepath, 'r', encoding='utf-8-sig') as csvfile:
            data = csv.DictReader(csvfile)
            self._import_cryptocom_swap(data)
            # reset the iterator
            csvfile.seek(0)
            # pass the header since seek(0) make the first row to be the header
            next(data)
            for row in data:
                try:
                    self._consume_cryptocom_entry(row)
                except UnknownAsset as e:
                    self.db.msg_aggregator.add_warning(
                        f'During cryptocom CSV import found action with unknown '
                        f'asset {e.asset_name}. Ignoring entry',
                    )
                    continue
                except DeserializationError as e:
                    self.db.msg_aggregator.add_warning(
                        f'Error during cryptocom CSV import deserialization. '
                        f'Error was {str(e)}. Ignoring entry',
                    )
                    continue
                except UnsupportedCryptocomEntry as e:
                    self.db.msg_aggregator.add_warning(str(e))
                    continue
        return None



import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Set, Tuple

from eth_utils.address import to_checksum_address

from rotkehlchen.accounting.structures import Balance
from rotkehlchen.assets.asset import Asset, EthereumToken
from rotkehlchen.chain.ethereum.aave.common import (
    AAVE_RESERVE_TO_ASSET,
    ASSET_TO_AAVE_RESERVE_ADDRESS,
    AaveBalances,
    AaveHistory,
    AaveInquirer,
    _get_reserve_address_decimals,
)
from rotkehlchen.chain.ethereum.graph import Graph
from rotkehlchen.chain.ethereum.makerdao.common import RAY
from rotkehlchen.chain.ethereum.structures import (
    AaveBorrowEvent,
    AaveEvent,
    AaveLiquidationEvent,
    AaveRepayEvent,
    AaveSimpleEvent,
)
from rotkehlchen.chain.ethereum.utils import token_normalized_value_decimals
from rotkehlchen.constants.ethereum import ATOKEN_ABI
from rotkehlchen.constants.misc import ZERO
from rotkehlchen.errors import UnknownAsset
from rotkehlchen.fval import FVal
from rotkehlchen.history.price import query_usd_price_zero_if_error
from rotkehlchen.inquirer import Inquirer
from rotkehlchen.typing import ChecksumEthAddress, Timestamp
from rotkehlchen.user_messages import MessagesAggregator
from rotkehlchen.utils.misc import ts_now

if TYPE_CHECKING:
    from rotkehlchen.chain.ethereum.manager import EthereumManager
    from rotkehlchen.db.dbhandler import DBHandler

log = logging.getLogger(__name__)

AAVE_GRAPH_RECENT_SECS = 600  # 10 mins

USER_RESERVES_QUERY = """
{{
  userReserves(where: {{ user: "{address}"}}) {{
    id
    reserve{{
      id
      symbol
    }}
    user {{
      id
    }}
  }}
}}"""


DEPOSIT_EVENTS_QUERY = """
  deposits (orderBy: timestamp, orderDirection: asc, where: {
   user: $address, timestamp_lte: $end_ts, timestamp_gte: $start_ts
  }) {
    id
    amount
    referrer {
      id
    }
    reserve {
      id
    }
    timestamp
  }
}
"""

USER_EVENTS_QUERY = """
  users (where: {id: $address}) {
    id
    depositHistory {
        id
        amount
        reserve {
          id
        }
        timestamp
    }
    redeemUnderlyingHistory {
        id
        amount
        reserve {
          id
        }
        timestamp
    }
    borrowHistory {
        id
        amount
        reserve {
          id
        }
        timestamp
        borrowRate
        borrowRateMode
        accruedBorrowInterest
    }
    repayHistory {
        id
        amountAfterFee
        fee
        reserve {
          id
        }
        timestamp
    }
    liquidationCallHistory {
        id
        collateralAmount
        collateralReserve {
          id
        }
        principalAmount
        principalReserve {
          id
        }
        timestamp
    }
    reserves{
        id
        aTokenBalanceHistory {
          id
          balance
          userBalanceIndex
          interestRedirectionAddress
          redirectedBalance
          timestamp
        }
    }
  }
}
"""


class ATokenBalanceHistory(NamedTuple):
    reserve_address: ChecksumEthAddress
    balance: FVal
    tx_hash: str
    timestamp: Timestamp


class AaveUserReserve(NamedTuple):
    address: ChecksumEthAddress
    symbol: str


class AaveEventProcessingResult(NamedTuple):
    interest_events: List[AaveSimpleEvent]
    total_earned_interest: Dict[Asset, Balance]
    total_lost: Dict[Asset, Balance]
    total_earned_liquidations: Dict[Asset, Balance]


def _calculate_loss(
        borrow_actions: List[AaveEvent],
        balances: AaveBalances,
) -> Tuple[Dict[Asset, Balance], Dict[Asset, Balance]]:
    """Returns a tuple of mapping of losses due to liquidation/borrowing and
    earnings due to keeping the principal repaid by the liquidation"""
    borrow_actions.sort(key=lambda event: event.timestamp)
    historical_borrow_balances: Dict[Asset, FVal] = defaultdict(FVal)
    total_lost: Dict[Asset, Balance] = defaultdict(Balance)
    total_earned: Dict[Asset, Balance] = defaultdict(Balance)

    for b_action in borrow_actions:
        if b_action.event_type == 'borrow':
            historical_borrow_balances[b_action.asset] -= b_action.value.amount  # type: ignore
        elif b_action.event_type == 'repay':
            historical_borrow_balances[b_action.asset] += b_action.value.amount  # type: ignore
        elif b_action.event_type == 'liquidation':
            # At liquidation you lose the collateral asset
            total_lost[b_action.collateral_asset] += b_action.collateral_balance  # type: ignore  # noqa: E501
            # And your principal asset is repaid and you gain it
            historical_borrow_balances[b_action.principal_asset] += b_action.principal_balance.amount  # type: ignore # noqa: E501
            total_earned[b_action.principal_asset] += b_action.principal_balance  # type: ignore

    for b_asset, amount in historical_borrow_balances.items():
        borrow_balance = balances.borrowing.get(b_asset.identifier, None)
        if borrow_balance is not None:
            amount += borrow_balance.balance.amount

        usd_price = Inquirer().find_usd_price(b_asset)
        total_lost[b_asset] = Balance(
            # add total_lost amount in case of liquidations
            amount=total_lost[b_asset].amount + amount,
            usd_value=amount * usd_price,
        )

    return total_lost, total_earned


def _parse_common_event_data(
        entry: Dict[str, Any],
        from_ts: Timestamp,
        to_ts: Timestamp,
) -> Optional[Tuple[Timestamp, str, int]]:
    """Parses and returns the common data of each event.

    Returns None if timestamp is out of range or if there is an error
    """
    timestamp = entry['timestamp']
    if timestamp < from_ts or timestamp > to_ts:
        # Since for the user data we can't query per timestamp, filter timestamps here
        return None

    pair = entry['id'].split(':')
    if len(pair) != 2:
        log.error(
            f'Could not parse the id entry for an aave liquidation as '
            f'returned by graph: {entry["id"]}.  Skipping entry ...',
        )
        return None

    tx_hash = pair[0]
    index = int(pair[1])  # not really log index
    return timestamp, tx_hash, index


def _parse_atoken_balance_history(
        history: List[Dict[str, Any]],
        from_ts: Timestamp,
        to_ts: Timestamp,
) -> List[ATokenBalanceHistory]:
    result = []
    for entry in history:
        timestamp = entry['timestamp']
        if timestamp < from_ts or timestamp > to_ts:
            continue

        entry_id = entry['id']
        pairs = entry_id.split('0x')
        if len(pairs) != 4:
            log.error(
                f'Expected to find 3 hashes in graps\'s aTokenBalanceHistory '
                f'id but the encountered id does not match: {entry_id}. Skipping entry...',
            )
            continue

        reserve_address = to_checksum_address('0x' + pairs[2])
        tx_hash = '0x' + pairs[3]
        asset = AAVE_RESERVE_TO_ASSET.get(reserve_address, None)
        if asset is None:
            log.error(
                f'Unknown aave reserve address returned by atoken balance history '
                f' graph query: {reserve_address}. Skipping entry ...',
            )
            continue

        _, decimals = _get_reserve_address_decimals(asset.identifier)
        balance = token_normalized_value_decimals(int(entry['balance']), token_decimals=decimals)
        result.append(ATokenBalanceHistory(
            reserve_address=reserve_address,
            balance=balance,
            tx_hash=tx_hash,
            timestamp=timestamp,
        ))

    return result


def _get_reserve_asset_and_decimals(
        entry: Dict[str, Any],
        reserve_key: str,
) -> Optional[Tuple[Asset, int]]:
    reserve_address = to_checksum_address(entry[reserve_key]['id'])
    asset = AAVE_RESERVE_TO_ASSET.get(reserve_address, None)
    if asset is None:
        log.error(
            f'Unknown aave reserve address returned by graph query: '
            f'{reserve_address}. Skipping entry ...',
        )
        return None

    _, decimals = _get_reserve_address_decimals(asset.identifier)
    return asset, decimals


class AaveGraphInquirer(AaveInquirer):
    """Reads Aave historical data from the graph protocol"""

    def __init__(
            self,
            ethereum_manager: 'EthereumManager',
            database: 'DBHandler',
            msg_aggregator: MessagesAggregator,
    ) -> None:
        """May raise RemoteError if we can't connect to the subgraph"""
        super().__init__(
            ethereum_manager=ethereum_manager,
            database=database,
            msg_aggregator=msg_aggregator,
        )
        self.graph = Graph('https://api.thegraph.com/subgraphs/name/aave/protocol-raw')

    def get_history_for_addresses(
            self,
            addresses: List[ChecksumEthAddress],
            to_block: int,
            from_timestamp: Timestamp,
            to_timestamp: Timestamp,
            aave_balances: Dict[ChecksumEthAddress, AaveBalances],
    ) -> Dict[ChecksumEthAddress, AaveHistory]:
        """
        Queries aave history for a list of addresses.

        This function should be entered while holding the history_lock
        semaphore
        """
        result = {}
        for address in addresses:
            history_results = self.get_history_for_address(
                user_address=address,
                from_timestamp=from_timestamp,
                to_timestamp=to_timestamp,
                balances=aave_balances.get(address, AaveBalances({}, {})),
            )
            if history_results is None:
                continue
            result[address] = history_results

        return result

    def _get_user_reserves(self, address: ChecksumEthAddress) -> List[AaveUserReserve]:
        query = self.graph.query(
            querystr=USER_RESERVES_QUERY.format(address=address.lower()),
        )
        result = []
        for entry in query['userReserves']:
            reserve = entry['reserve']
            result.append(AaveUserReserve(
                address=to_checksum_address(reserve['id']),
                symbol=reserve['symbol'],
            ))

        return result

    def _calculate_interest_and_profit(
            self,
            user_address: ChecksumEthAddress,
            user_result: Dict[str, Any],
            actions: List[AaveSimpleEvent],
            balances: AaveBalances,
            db_interest_events: Set[AaveSimpleEvent],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> Tuple[List[AaveSimpleEvent], Dict[Asset, Balance]]:
        reserve_history = {}
        for reserve in user_result['reserves']:
            pairs = reserve['id'].split('0x')
            if len(pairs) != 3:
                log.error(
                    f'Expected to find 2 hashes in graph\'s reserve history id '
                    f'but the encountered id does not match: {reserve["id"]}. Skipping entry...',
                )
            reserve_address = to_checksum_address('0x' + pairs[2])
            atoken_history = _parse_atoken_balance_history(
                history=reserve['aTokenBalanceHistory'],
                from_ts=from_ts,
                to_ts=to_ts,
            )
            reserve_history[reserve_address] = atoken_history

        interest_events: List[AaveSimpleEvent] = []
        atoken_balances: Dict[Asset, FVal] = defaultdict(FVal)
        used_history_indices = set()
        total_earned: Dict[Asset, Balance] = defaultdict(Balance)

        # Go through the existing db interest events and add total earned
        for interest_event in db_interest_events:
            total_earned[interest_event.asset] += interest_event.value

        # Create all new interest events in the query
        actions.sort(key=lambda event: event.timestamp)
        for action in actions:
            if action.event_type == 'deposit':
                atoken_balances[action.asset] += action.value.amount
            else:  # withdrawal
                atoken_balances[action.asset] -= action.value.amount

            reserve_address = ASSET_TO_AAVE_RESERVE_ADDRESS.get(action.asset.identifier, None)  # type: ignore  # noqa: E501
            history = reserve_history.get(reserve_address, None)
            if history is None:
                log.error(
                    f'Could not find aTokenBalanceHistory for reserve '
                    f'{reserve_address} in an aave graph response.'
                    f' Skipping entry...',
                )
                continue
            history.sort(key=lambda event: event.timestamp)

            for idx, entry in enumerate(history):
                if idx in used_history_indices:
                    continue
                used_history_indices.add(idx)

                if entry.tx_hash == action.tx_hash:
                    diff = entry.balance - atoken_balances[action.asset]
                    if diff != ZERO:
                        atoken_balances[action.asset] = entry.balance
                        try:
                            asset = EthereumToken('a' + action.asset.identifier)
                        except UnknownAsset:
                            log.error(
                                f'Could not find corresponding aToken to'
                                f'{action.asset.identifier} during an aave graph uery'
                                f' Skipping entry...',
                            )
                            continue
                        timestamp = entry.timestamp
                        usd_price = query_usd_price_zero_if_error(
                            asset=asset,
                            time=timestamp,
                            location='aave interest event from graph query',
                            msg_aggregator=self.msg_aggregator,
                        )
                        earned_balance = Balance(amount=diff, usd_value=diff * usd_price)
                        interest_event = AaveSimpleEvent(
                            event_type='interest',
                            asset=asset,
                            value=earned_balance,
                            block_number=0,  # can't get from graph query
                            timestamp=timestamp,
                            tx_hash=entry.tx_hash,
                            # not really the log index, but should also be unique
                            log_index=action.log_index + 1,
                        )
                        if interest_event in db_interest_events:
                            # This should not really happen since we already query
                            # historical atoken balance history in the new range
                            log.warning(
                                f'During aave subgraph query interest and profit calculation '
                                f'tried to generate interest event {interest_event} that '
                                f'already existed in the DB ',
                            )
                            continue

                        interest_events.append(interest_event)
                        total_earned[asset] += earned_balance

                    # and once done break off the loop
                    break

                else:
                    # this atoken history is not due to an action, so skip it
                    # it's probably due to a simple transfer
                    atoken_balances[action.asset] = entry.balance
                    if action.event_type == 'deposit':
                        atoken_balances[action.asset] += action.value.amount
                    else:  # withdrawal
                        atoken_balances[action.asset] -= action.value.amount

        # Take aave unpaid interest into account
        for symbol, lending_balance in balances.lending.items():
            atoken = EthereumToken('a' + symbol)
            principal_balance = self.ethereum.call_contract(
                contract_address=atoken.ethereum_address,
                abi=ATOKEN_ABI,
                method_name='principalBalanceOf',
                arguments=[user_address],
            )
            unpaid_interest = lending_balance.balance.amount - (principal_balance / (FVal(10) ** FVal(atoken.decimals)))  # noqa: E501
            usd_price = Inquirer().find_usd_price(atoken)
            total_earned[atoken] += Balance(
                amount=unpaid_interest,
                usd_value=unpaid_interest * usd_price,
            )

        return interest_events, total_earned

    def _process_events(
            self,
            user_address: ChecksumEthAddress,
            user_result: Dict[str, Any],
            from_ts: Timestamp,
            to_ts: Timestamp,
            deposits: List[AaveSimpleEvent],
            withdrawals: List[AaveSimpleEvent],
            borrows: List[AaveBorrowEvent],
            repays: List[AaveRepayEvent],
            liquidations: List[AaveLiquidationEvent],
            db_events: List[AaveEvent],
            balances: AaveBalances,
    ) -> AaveEventProcessingResult:
        """Calculates the interest events and the total earned from all the given events.
        Also calculates total loss from borrowing and liquidations.

        Also returns the edited DB events
        """
        actions: List[AaveSimpleEvent] = []
        borrow_actions: List[AaveEvent] = []
        db_interest_events: Set[AaveSimpleEvent] = set()
        for db_event in db_events:
            if db_event.event_type == 'deposit':
                actions.append(db_event)  # type: ignore
            elif db_event.event_type == 'withdrawal':
                actions.append(db_event)  # type: ignore
            elif db_event.event_type == 'interest':
                db_interest_events.add(db_event)  # type: ignore
            elif db_event.event_type == 'borrow':
                borrow_actions.append(db_event)
            elif db_event.event_type == 'repay':
                borrow_actions.append(db_event)
            elif db_event.event_type == 'liquidation':
                borrow_actions.append(db_event)

        interest_events, total_earned = self._calculate_interest_and_profit(
            user_address=user_address,
            user_result=user_result,
            actions=actions + deposits + withdrawals,
            balances=balances,
            db_interest_events=db_interest_events,
            from_ts=from_ts,
            to_ts=to_ts,
        )
        total_lost, total_earned_liquidations = _calculate_loss(
            borrow_actions=borrow_actions + borrows + repays + liquidations,  # type: ignore
            balances=balances,
        )
        return AaveEventProcessingResult(
            interest_events=interest_events,
            total_earned_interest=total_earned,
            total_lost=total_lost,
            total_earned_liquidations=total_earned_liquidations,
        )

    def _get_user_data(
            self,
            from_ts: Timestamp,
            to_ts: Timestamp,
            address: ChecksumEthAddress,
            balances: AaveBalances,
    ) -> AaveHistory:
        last_query = self.database.get_used_query_range(f'aave_events_{address}')
        db_events = self.database.get_aave_events(address=address)

        now = ts_now()
        last_query_ts = 0
        if last_query is not None:
            last_query_ts = last_query[1]
            from_ts = Timestamp(last_query_ts + 1)

        deposits = withdrawals = borrows = repays = liquidation_calls = []
        query = self.graph.query(
            querystr=USER_EVENTS_QUERY,
            param_types={'$address': 'ID!'},
            param_values={'address': address.lower()},
        )
        user_result = query['users'][0]
        if now - last_query_ts > AAVE_GRAPH_RECENT_SECS:
            # In theory if these were individual queries we should do them only if
            # we have not queried recently. In practise since we only do 1 query above
            # this is useless for now, but keeping the mechanism in case we change
            # the way we query the subgraph
            deposits = self._parse_deposits(user_result['depositHistory'], from_ts, to_ts)
            withdrawals = self._parse_withdrawals(
                withdrawals=user_result['redeemUnderlyingHistory'],
                from_ts=from_ts,
                to_ts=to_ts,
            )
            borrows = self._parse_borrows(user_result['borrowHistory'], from_ts, to_ts)
            repays = self._parse_repays(user_result['repayHistory'], from_ts, to_ts)
            liquidation_calls = self._parse_liquidations(
                user_result['liquidationCallHistory'],
                from_ts,
                to_ts,
            )

        result = self._process_events(
            user_address=address,
            user_result=user_result,
            from_ts=from_ts,
            to_ts=to_ts,
            deposits=deposits,
            withdrawals=withdrawals,
            borrows=borrows,
            repays=repays,
            liquidations=liquidation_calls,
            db_events=db_events,
            balances=balances,
        )

        # Add all new events to the DB
        new_events: List[AaveEvent] = deposits + withdrawals + result.interest_events + borrows + repays + liquidation_calls  # type: ignore  # noqa: E501
        self.database.add_aave_events(address, new_events)
        # After all events have been queried then also update the query range.
        # Even if no events are found for an address we need to remember the range
        self.database.update_used_query_range(
            name=f'aave_events_{address}',
            start_ts=Timestamp(0),
            end_ts=now,
        )

        # Sort actions so that actions with same time are sorted deposit -> interest -> withdrawal
        all_events: List[AaveEvent] = new_events + db_events
        sort_map = {'deposit': 0, 'interest': 0.1, 'withdrawal': 0.2, 'borrow': 0.3, 'repay': 0.4, 'liquidation': 0.5}  # noqa: E501
        all_events.sort(key=lambda event: sort_map[event.event_type] + event.timestamp)
        return AaveHistory(
            events=all_events,
            total_earned_interest=result.total_earned_interest,
            total_lost=result.total_lost,
            total_earned_liquidations=result.total_earned_liquidations,
        )

    def _parse_deposits(
            self,
            deposits: List[Dict[str, Any]],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> List[AaveSimpleEvent]:
        events: List[AaveSimpleEvent] = []
        for entry in deposits:
            common = _parse_common_event_data(entry, from_ts, to_ts)
            if common is None:
                continue  # either timestamp out of range or error (logged in the function above)
            timestamp, tx_hash, index = common
            result = self._get_asset_and_balance(
                entry=entry,
                timestamp=timestamp,
                reserve_key='reserve',
                amount_key='amount',
                location='aave deposit from graph query',
            )
            if result is None:
                continue  # problem parsing, error already logged
            asset, balance = result
            events.append(AaveSimpleEvent(
                event_type='deposit',
                asset=asset,
                value=balance,
                block_number=0,  # can't get from graph query
                timestamp=timestamp,
                tx_hash=tx_hash,
                log_index=index,  # not really the log index, but should also be unique
            ))

        return events

    def _parse_withdrawals(
            self,
            withdrawals: List[Dict[str, Any]],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> List[AaveSimpleEvent]:
        events = []
        for entry in withdrawals:
            common = _parse_common_event_data(entry, from_ts, to_ts)
            if common is None:
                continue  # either timestamp out of range or error (logged in the function above)
            timestamp, tx_hash, index = common
            result = self._get_asset_and_balance(
                entry=entry,
                timestamp=timestamp,
                reserve_key='reserve',
                amount_key='amount',
                location='aave withdrawal from graph query',
            )
            if result is None:
                continue  # problem parsing, error already logged
            asset, balance = result
            events.append(AaveSimpleEvent(
                event_type='withdrawal',
                asset=asset,
                value=balance,
                block_number=0,  # can't get from graph query
                timestamp=timestamp,
                tx_hash=tx_hash,
                log_index=index,  # not really the log index, but should also be unique
            ))

        return events

    def _parse_borrows(
            self,
            borrows: List[Dict[str, Any]],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> List[AaveBorrowEvent]:
        events = []
        for entry in borrows:
            common = _parse_common_event_data(entry, from_ts, to_ts)
            if common is None:
                continue  # either timestamp out of range or error (logged in the function above)
            timestamp, tx_hash, index = common
            result = self._get_asset_and_balance(
                entry=entry,
                timestamp=timestamp,
                reserve_key='reserve',
                amount_key='amount',
                location='aave borrow from graph query',
            )
            if result is None:
                continue  # problem parsing, error already logged
            asset, balance = result
            borrow_rate = FVal(entry['borrowRate']) / RAY
            borrow_rate_mode = entry['borrowRateMode']
            accrued_borrow_interest = entry['accruedBorrowInterest']
            events.append(AaveBorrowEvent(
                event_type='borrow',
                asset=asset,
                value=balance,
                borrow_rate_mode=borrow_rate_mode.lower(),
                borrow_rate=borrow_rate,
                accrued_borrow_interest=accrued_borrow_interest,
                block_number=0,  # can't get from graph query
                timestamp=timestamp,
                tx_hash=tx_hash,
                log_index=index,  # not really the log index, but should also be unique
            ))

        return events

    def _parse_repays(
            self,
            repays: List[Dict[str, Any]],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> List[AaveRepayEvent]:
        events = []
        for entry in repays:
            common = _parse_common_event_data(entry, from_ts, to_ts)
            if common is None:
                continue  # either timestamp out of range or error (logged in the function above)
            timestamp, tx_hash, index = common
            result = _get_reserve_asset_and_decimals(entry, reserve_key='reserve')
            if result is None:
                continue  # problem parsing, error already logged
            asset, decimals = result
            amount_after_fee = token_normalized_value_decimals(
                int(entry['amountAfterFee']),
                token_decimals=decimals,
            )
            fee = token_normalized_value_decimals(int(entry['fee']), token_decimals=decimals)
            usd_price = query_usd_price_zero_if_error(
                asset=asset,
                time=timestamp,
                location='aave repay from graph query',
                msg_aggregator=self.msg_aggregator,
            )
            events.append(AaveRepayEvent(
                event_type='repay',
                asset=asset,
                value=Balance(amount=amount_after_fee, usd_value=amount_after_fee * usd_price),
                fee=Balance(amount=fee, usd_value=fee * usd_price),
                block_number=0,  # can't get from graph query
                timestamp=timestamp,
                tx_hash=tx_hash,
                log_index=index,  # not really the log index, but should also be unique
            ))

        return events

    def _parse_liquidations(
            self,
            liquidations: List[Dict[str, Any]],
            from_ts: Timestamp,
            to_ts: Timestamp,
    ) -> List[AaveLiquidationEvent]:
        events = []
        for entry in liquidations:
            common = _parse_common_event_data(entry, from_ts, to_ts)
            if common is None:
                continue  # either timestamp out of range or error (logged in the function above)
            timestamp, tx_hash, index = common
            result = self._get_asset_and_balance(
                entry=entry,
                timestamp=timestamp,
                reserve_key='collateralReserve',
                amount_key='collateralAmount',
                location='aave liquidation from graph query',
            )
            if result is None:
                continue  # problem parsing, error already logged
            collateral_asset, collateral_balance = result

            result = self._get_asset_and_balance(
                entry=entry,
                timestamp=timestamp,
                reserve_key='principalReserve',
                amount_key='principalAmount',
                location='aave liquidation from graph query',
            )
            if result is None:
                continue  # problem parsing, error already logged
            principal_asset, principal_balance = result
            events.append(AaveLiquidationEvent(
                event_type='liquidation',
                collateral_asset=collateral_asset,
                collateral_balance=collateral_balance,
                principal_asset=principal_asset,
                principal_balance=principal_balance,
                block_number=0,  # can't get from graph query
                timestamp=timestamp,
                tx_hash=tx_hash,
                log_index=index,  # not really the log index, but should also be unique
            ))

        return events

    def get_history_for_address(
            self,
            user_address: ChecksumEthAddress,
            from_timestamp: Timestamp,
            to_timestamp: Timestamp,
            balances: AaveBalances,
    ) -> Optional[AaveHistory]:
        """
        Queries aave history for a single address.

        This function should be entered while holding the history_lock
        semaphore
        """
        reserves = self._get_user_reserves(address=user_address)
        if len(reserves) != 0:
            return self._get_user_data(
                from_ts=from_timestamp,
                to_ts=to_timestamp,
                address=user_address,
                balances=balances,
            )

        return None

    def _get_asset_and_balance(
            self,
            entry: Dict[str, Any],
            timestamp: Timestamp,
            reserve_key: str,
            amount_key: str,
            location: str,
    ) -> Optional[Tuple[Asset, Balance]]:
        """Utility function to parse asset from graph query amount and price and return balance"""
        result = _get_reserve_asset_and_decimals(entry, reserve_key)
        if result is None:
            return None
        asset, decimals = result
        amount = token_normalized_value_decimals(
            token_amount=int(entry[amount_key]),
            token_decimals=decimals,
        )
        usd_price = query_usd_price_zero_if_error(
            asset=asset,
            time=timestamp,
            location=location,
            msg_aggregator=self.msg_aggregator,
        )
        return asset, Balance(amount=amount, usd_value=amount * usd_price)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rotkehlchen.db.dbhandler import DBHandler


def upgrade_v18_to_v19(db: 'DBHandler') -> None:
    """Upgrades the DB from v18 to v19

    - Deletes all aave historical data and cache since they are going to be requeried via the graph
    - Deletes and recreates the table with the new schema
    https://github.com/rotki/rotki/issues/1494
    - Deletes all poloniex trades/deposit/withdrawals so they can be requeried after #1631
    """
    cursor = db.conn.cursor()
    cursor.execute('DELETE FROM aave_events;')
    cursor.execute('DELETE FROM used_query_ranges WHERE name LIKE "aave_events%";')
    cursor.execute('DROP TABLE IF EXISTS aave_events;')
    cursor.execute("""
CREATE TABLE IF NOT EXISTS aave_events (
    address VARCHAR[42] NOT NULL,
    event_type VARCHAR[10] NOT NULL,
    block_number INTEGER NOT NULL,
    timestamp INTEGER NOT NULL,
    tx_hash VARCHAR[66] NOT NULL,
    log_index INTEGER NOT NULL,
    asset1 VARCHAR[12] NOT NULL,
    asset1_amount TEXT NOT NULL,
    asset1_usd_value TEXT NOT NULL,
    asset2 VARCHAR[12],
    asset2amount_borrowrate_feeamount TEXT,
    asset2usd_value_accruedinterest_feeusdvalue TEXT,
    borrow_rate_mode VARCHAR[10],
    PRIMARY KEY (event_type, tx_hash, log_index)
);
""")
    # Delete all poloniex trades/deposits/withdrawals so they can be requeried
    cursor.execute('DELETE FROM trades WHERE location="C";')
    cursor.execute('DELETE FROM asset_movements WHERE location="C";')
    cursor.execute('DELETE FROM used_query_ranges WHERE name LIKE "poloniex_%";')
    db.conn.commit()



from typing import TYPE_CHECKING, Dict, Optional

from eth_utils.address import to_checksum_address

from rotkehlchen.constants.ethereum import MAKERDAO_PROXY_REGISTRY
from rotkehlchen.db.dbhandler import DBHandler
from rotkehlchen.typing import ChecksumEthAddress
from rotkehlchen.user_messages import MessagesAggregator
from rotkehlchen.utils.interfaces import EthereumModule
from rotkehlchen.utils.misc import ts_now

if TYPE_CHECKING:
    from rotkehlchen.chain.ethereum.manager import EthereumManager

MAKERDAO_REQUERY_PERIOD = 7200  # Refresh queries every 2 hours

WAD_DIGITS = 18
WAD = 10**WAD_DIGITS
RAY_DIGITS = 27
RAY = 10**RAY_DIGITS
RAD_DIGITS = 45
RAD = 10**RAD_DIGITS


class MakerDAOCommon(EthereumModule):
    """Class to manage MakerDAO related stuff such as DSR and CDPs/Vaults"""

    def __init__(
            self,
            ethereum_manager: 'EthereumManager',
            database: DBHandler,
            msg_aggregator: MessagesAggregator,
    ) -> None:
        self.ethereum = ethereum_manager
        self.database = database
        self.msg_aggregator = msg_aggregator
        self.proxy_mappings: Dict[ChecksumEthAddress, ChecksumEthAddress] = {}
        self.reset_last_query_ts()

    def reset_last_query_ts(self) -> None:
        """Reset the last query timestamps, effectively cleaning the caches"""
        self.last_proxy_mapping_query_ts = 0

    def _get_account_proxy(self, address: ChecksumEthAddress) -> Optional[ChecksumEthAddress]:
        """Checks if a DSR proxy exists for the given address and returns it if it does

        May raise:
        - RemoteError if etherscan is used and there is a problem with
        reaching it or with the returned result.
        - BlockchainQueryError if an ethereum node is used and the contract call
        queries fail for some reason
        """
        result = MAKERDAO_PROXY_REGISTRY.call(self.ethereum, 'proxies', arguments=[address])
        if int(result, 16) != 0:
            return to_checksum_address(result)
        return None

    def _get_accounts_having_maker_proxy(self) -> Dict[ChecksumEthAddress, ChecksumEthAddress]:
        """Returns a mapping of accounts that have DSR proxies to their proxies

        If the proxy mappings have been queried in the past REQUERY_PERIOD
        seconds then the old result is used.

        May raise:
        - RemoteError if etherscan is used and there is a problem with
        reaching it or with the returned result.
        - BlockchainQueryError if an ethereum node is used and the contract call
        queries fail for some reason
        """
        now = ts_now()
        if now - self.last_proxy_mapping_query_ts < MAKERDAO_REQUERY_PERIOD:
            return self.proxy_mappings

        mapping = {}
        accounts = self.database.get_blockchain_accounts()
        for account in accounts.eth:
            proxy_result = self._get_account_proxy(account)
            if proxy_result:
                mapping[account] = proxy_result

        self.last_proxy_mapping_query_ts = ts_now()
        self.proxy_mappings = mapping
        return mapping

    # -- Methods following the EthereumModule interface -- #
    def on_startup(self) -> None:
        pass

    def on_account_addition(self, address: ChecksumEthAddress) -> None:
        self.reset_last_query_ts()
        # Get the proxy of the account
        proxy_result = self._get_account_proxy(address)
        if proxy_result is None:
            return

        # add it to the mapping
        self.proxy_mappings[address] = proxy_result

    def on_account_removal(self, address: ChecksumEthAddress) -> None:
        self.reset_last_query_ts()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rotkehlchen.db.dbhandler import DBHandler


def _upgrade_blockchain_accounts_table(db: 'DBHandler') -> None:
    cursor = db.conn.cursor()
    # This is the user credentiaals trades table at v10
    query = cursor.execute('SELECT blockchain, account from blockchain_accounts;')
    account_tuples = []
    for result in query:
        account_tuples.append((result[0], result[1], None))

    # We got all data. Now delete the old table and create the new one
    cursor.execute('DROP TABLE blockchain_accounts;')
    db.conn.commit()
    # Now create blockchain accounts table at v11
    cursor.execute("""CREATE TABLE IF NOT EXISTS blockchain_accounts (
    blockchain VARCHAR[24],
    account TEXT NOT NULL PRIMARY KEY,
    label TEXT
);""")
    db.conn.commit()
    # and finally move the data to the new table
    cursor.executemany(
        'INSERT INTO blockchain_accounts(blockchain, account, label) '
        'VALUES (?, ?, ?)',
        account_tuples,
    )
    db.conn.commit()


def upgrade_v10_to_v11(db: 'DBHandler') -> None:
    """Upgrades the DB from v10 to v11

    - Adds the label column to blockchain accounts
    """
    _upgrade_blockchain_accounts_table(db)



from enum import Enum
from sqlite3 import Cursor
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Tuple, Union

from typing_extensions import Literal

from rotkehlchen.accounting.structures import BalanceType
from rotkehlchen.assets.asset import Asset
from rotkehlchen.balances.manual import ManuallyTrackedBalance
from rotkehlchen.typing import (
    BlockchainAccountData,
    BTCAddress,
    ChecksumEthAddress,
    HexColorCode,
    ListOfBlockchainAddresses,
    SupportedBlockchain,
    Timestamp,
)
from rotkehlchen.utils.misc import rgetattr

if TYPE_CHECKING:
    from rotkehlchen.chain.bitcoin.xpub import XpubData


class BlockchainAccounts(NamedTuple):
    eth: List[ChecksumEthAddress]
    btc: List[BTCAddress]

    def get(self, blockchain: SupportedBlockchain) -> ListOfBlockchainAddresses:
        if blockchain == SupportedBlockchain.ETHEREUM:
            return self.eth

        return self.btc


class AssetBalance(NamedTuple):
    category: BalanceType
    time: Timestamp
    asset: Asset
    amount: str
    usd_value: str


class SingleAssetBalance(NamedTuple):
    category: BalanceType
    time: Timestamp
    amount: str
    usd_value: str


class LocationData(NamedTuple):
    time: Timestamp
    location: str  # Location serialized in a DB enum
    usd_value: str


class Tag(NamedTuple):
    name: str
    description: Optional[str]
    background_color: HexColorCode
    foreground_color: HexColorCode

    def serialize(self) -> Dict[str, str]:
        return self._asdict()  # pylint: disable=no-member


class DBStartupAction(Enum):
    NOTHING = 1
    UPGRADE_3_4 = 2
    STUCK_4_3 = 3


def str_to_bool(s: str) -> bool:
    return True if s == 'True' else False


def form_query_to_filter_timestamps(
        query: str,
        timestamp_attribute: str,
        from_ts: Optional[Timestamp],
        to_ts: Optional[Timestamp],
) -> Tuple[str, Union[Tuple, Tuple[Timestamp], Tuple[Timestamp, Timestamp]]]:
    """Formulates the query string and its bindings to filter for timestamps"""
    bindings: Union[Tuple, Tuple[Timestamp], Tuple[Timestamp, Timestamp]] = ()
    got_from_ts = from_ts is not None
    got_to_ts = to_ts is not None
    if (got_from_ts or got_to_ts):
        if 'WHERE' not in query:
            query += 'WHERE '
        else:
            query += 'AND '

    if got_from_ts:
        query += f'{timestamp_attribute} >= ? '
        bindings = (from_ts,)
        if got_to_ts:
            query += f'AND {timestamp_attribute} <= ? '
            bindings = (from_ts, to_ts)
    elif got_to_ts:
        query += f'AND {timestamp_attribute} <= ? '
        bindings = (to_ts,)

    query += f'ORDER BY {timestamp_attribute} ASC;'
    return query, bindings


def deserialize_tags_from_db(val: Optional[str]) -> Optional[List[str]]:
    """Read tags from the DB and turn it into a List of tags"""
    if val is None:
        tags = None
    else:
        tags = val.split(',')
        if len(tags) == 1 and tags[0] == '':
            tags = None

    return tags


def insert_tag_mappings(
        cursor: Cursor,
        data: Union[List[ManuallyTrackedBalance], List[BlockchainAccountData], List['XpubData']],
        object_reference_keys: List[
            Literal['label', 'address', 'xpub.xpub', 'derivation_path'],
        ],
) -> None:
    """
    Inserts the tag mappings from a list of potential data entries. If multiple keys are given
    then the concatenation of their values is what goes in as the object reference.
    """
    mapping_tuples = []
    for entry in data:
        if entry.tags is not None:
            reference = ''
            for key in object_reference_keys:
                value = rgetattr(entry, key)
                if value is not None:
                    reference += value
            mapping_tuples.extend([(reference, tag) for tag in entry.tags])

    cursor.executemany(
        'INSERT INTO tag_mappings(object_reference, tag_name) VALUES (?, ?)', mapping_tuples,
    )



"""
Assets removed from all_assets.json in https://github.com/rotki/rotki/pull/1354
due to no price support neither in cryptocompare nor in coingecko
"""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rotkehlchen.db.dbhandler import DBHandler


REMOVED_ASSETS = [
    'ACC-2',
    'APH-2',
    'ATOM-2',
    'BLZ-2',
    'FORK',
    'KNC-2',
    'LUNA',
    'MARS',
    'POLY-2',
    'PRC-2',
    'TALK',
    'USDS-2',
]
REMOVED_ETH_TOKENS = [
    'AI',
    'BLX',
    'BNN',
    'BTK',
    'CAT-2',
    'DAXT',
    'DEW',
    'E4ROW',
    'EDU-2',
    'EXC',
    'FID',
    'FLX',
    'FMF',
    'FT-2',
    'JOT',
    'LCT-2',
    'LML',
    'MORE-2',
    'NCC-2',
    'NTO',
    'NXX',
    'OCC-2',
    'PASS-2',
    'RGS',
    'SMART-2',
    'SNIP',
    'STRC',
    'XMCT',
    'XPA',
]


def _delete_unsupported_assets(db: 'DBHandler') -> None:
    """Deletes all assets that are no longer supported from various DB tables"""
    cursor = db.conn.cursor()
    # Kind of heavy handed, but let's just delete the ethereum tokens cache
    # Easier query than deleting tokens from the string that is a json list
    cursor.execute('DELETE FROM ethereum_accounts_details;')
    db.conn.commit()
    tuples = [(x,) for x in REMOVED_ASSETS + REMOVED_ETH_TOKENS]
    # Delete all timed balances that contain any of the removed assets
    cursor.executemany(
        'DELETE FROM timed_balances WHERE currency=?;',
        tuples,
    )
    db.conn.commit()
    # Delete all manually tracked balances that contain any of the removed assets
    cursor.executemany(
        'DELETE FROM manually_tracked_balances WHERE asset=?;',
        tuples,
    )
    db.conn.commit()
    # Delete all ignored asset entries that may contain removed assets
    cursor.executemany(
        'DELETE FROM multisettings WHERE name=? AND value=?;',
        [('ignored_asset', x) for x in REMOVED_ASSETS + REMOVED_ETH_TOKENS],
    )
    db.conn.commit()
    # Delete all trades that contain a removed asset in either part of the pair or in the fee
    cursor.executemany(
        'DELETE FROM trades WHERE pair LIKE ? ESCAPE ?;',
        [(f'{x}\\_%', '\\') for x in REMOVED_ASSETS + REMOVED_ETH_TOKENS],
    )
    cursor.executemany(
        'DELETE FROM trades WHERE pair LIKE ? ESCAPE ?;',
        [(f'%\\_{x}', '\\') for x in REMOVED_ASSETS + REMOVED_ETH_TOKENS],
    )
    cursor.executemany(
        'DELETE FROM trades WHERE fee_currency = ?;',
        tuples,
    )
    db.conn.commit()
    # Delete all asset movements that deal with removed assets
    cursor.executemany(
        'DELETE FROM asset_movements WHERE asset=? OR fee_asset=?;',
        [(x, x) for x in REMOVED_ASSETS + REMOVED_ETH_TOKENS],
    )
    db.conn.commit()


def upgrade_v13_to_v14(db: 'DBHandler') -> None:
    """Upgrades the DB from v13 to v14

    - Deletes all assets that are no longer supported from the DB
    """
    _delete_unsupported_assets(db)



from pathlib import Path
from typing import TYPE_CHECKING

from rotkehlchen.crypto import sha3
from rotkehlchen.errors import DBUpgradeError

if TYPE_CHECKING:
    from rotkehlchen.db.dbhandler import DBHandler


def _remove_cache_files(user_data_dir: Path) -> None:
    """At 5->6 version upgrade all cache files should be removed

    That's since we moved all trades in the DB and as such it no longer makes
    any sense to have the cache files.
    """
    for p in user_data_dir.glob('*_trades.json'):
        try:
            p.unlink()
        except OSError:
            pass

    for p in user_data_dir.glob('*_history.json'):
        try:
            p.unlink()
        except OSError:
            pass

    for p in user_data_dir.glob('*_deposits_withdrawals.json'):
        try:
            p.unlink()
        except OSError:
            pass

    try:
        (user_data_dir / 'ethereum_tx_log.json').unlink()
    except OSError:
        pass


def _upgrade_trades_table(db: 'DBHandler') -> None:
    cursor = db.conn.cursor()
    # This is the data trades table at v5
    query = cursor.execute(
        """SELECT time, location, pair, type, amount, rate, fee, fee_currency,
        link, notes FROM trades;""",
    )
    trade_tuples = []
    for result in query:
        # This is the logic of trade addition in v6 of the DB
        time = result[0]
        pair = result[2]
        old_trade_type = result[3]
        # hand deserialize trade type from DB enum since this code is going to stay
        # here even if deserialize_trade_type_from_db() changes
        if old_trade_type == 'buy':
            trade_type = 'A'
        elif old_trade_type == 'sell':
            trade_type = 'B'
        else:
            raise DBUpgradeError(
                f'Unexpected trade_type "{trade_type}" found while upgrading '
                f'from DB version 5 to 6',
            )

        trade_id = sha3(('external' + str(time) + str(old_trade_type) + pair).encode()).hex()
        trade_tuples.append((
            trade_id,
            time,
            'A',  # Symbolizes external in the location enum
            pair,
            trade_type,
            result[4],
            result[5],
            result[6],
            result[7],
            result[8],
            result[9],
        ))

    # We got all the external trades data. Now delete the old table and create
    # the new one
    cursor.execute('DROP TABLE trades;')
    db.conn.commit()
    # This is the scheme of the trades table at v6 from db/utils.py
    cursor.execute("""
    CREATE TABLE IF NOT EXISTS trades (
    id TEXT PRIMARY KEY,
    time INTEGER,
    location VARCHAR[24],
    pair VARCHAR[24],
    type CHAR(1) NOT NULL DEFAULT ('B') REFERENCES trade_type(type),
    amount TEXT,
    rate TEXT,
    fee TEXT,
    fee_currency VARCHAR[10],
    link TEXT,
    notes TEXT
    );""")
    db.conn.commit()

    # and finally move the data to the new table
    cursor.executemany(
        'INSERT INTO trades('
        '  id, '
        '  time,'
        '  location,'
        '  pair,'
        '  type,'
        '  amount,'
        '  rate,'
        '  fee,'
        '  fee_currency,'
        '  link,'
        '  notes)'
        'VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)',
        trade_tuples,
    )
    db.conn.commit()


def _location_to_enum_location(location: str) -> str:
    """Serialize location strings to DB location enums

    The reason we have a specialized function here and not just using
    deserialize_location(location).serialize_for_db() is that this code
    should work in the future if either of the two functions change or dissapear.
    """
    if location == 'external':
        return 'A'
    elif location == 'kraken':
        return 'B'
    elif location == 'poloniex':
        return 'C'
    elif location == 'bittrex':
        return 'D'
    elif location == 'binance':
        return 'E'
    elif location == 'bitmex':
        return 'F'
    elif location == 'coinbase':
        return 'G'
    elif location == 'total':
        return 'H'
    elif location == 'banks':
        return 'I'
    elif location == 'blockchain':
        return 'J'

    raise DBUpgradeError(f'Invalid location {location} encountered during DB v5->v6 upgrade')


def _upgrade_timed_location_data(db: 'DBHandler') -> None:
    cursor = db.conn.cursor()
    # This is the timed location data table at v5
    query = cursor.execute('SELECT time, location, usd_value FROM timed_location_data;')
    tuples = []
    for result in query:
        tuples.append((result[0], _location_to_enum_location(result[1]), result[2]))

    # We got all the old timed location data. Now delete the old table and create
    # the new one
    cursor.execute('DROP TABLE timed_location_data;')
    db.conn.commit()
    # This is the scheme of the timed_location_Data table at v6 from db/utils.py
    cursor.execute("""
    CREATE TABLE IF NOT EXISTS timed_location_data (
        time INTEGER,
        location CHAR(1) NOT NULL DEFAULT('A') REFERENCES location(location),
        usd_value TEXT,
        PRIMARY KEY (time, location)
    );
    """)
    db.conn.commit()

    # and finally move the data to the new table
    cursor.executemany(
        'INSERT INTO timed_location_data(time, location, usd_value)'
        'VALUES (?, ?, ?)',
        tuples,
    )
    db.conn.commit()


def upgrade_v5_to_v6(db: 'DBHandler') -> None:
    """Upgrades the DB from v5 to v6

    - removes all cache files
    - upgrades trades table to use the new id scheme
    - upgrades trades table to use an enum table for trade type
    - upgrades trades table to use an enum table for location
    - upgrades timed_location_data table to use an enum table for location
    """
    _remove_cache_files(db.user_data_dir)
    _upgrade_trades_table(db)
    _upgrade_timed_location_data(db)




from .typing import (  # noqa: F401
    UNISWAP_EVENTS_PREFIX,
    UNISWAP_TRADES_PREFIX,
    LiquidityPool as UniswapPool,
    LiquidityPoolAsset as UniswapPoolAsset,
    LiquidityPoolEvent as UniswapPoolEvent,
    LiquidityPoolEventsBalance as UniswapPoolEventsBalance,
)
from .uniswap import Uniswap  # noqa: F401

__all__ = [
    'UNISWAP_EVENTS_PREFIX',
    'UNISWAP_TRADES_PREFIX',
    'UniswapPool',
    'UniswapPoolAsset',
    'UniswapPoolEvent',
    'UniswapPoolEventsBalance',
    'Uniswap',
]



# -*- coding: utf-8 -*-

import pandas, hashlib, threading
import koalafolio.web.cryptocompareApi as ccapi
import koalafolio.PcpCore.settings as settings
import koalafolio.Import.Converter as converter
import koalafolio.PcpCore.logger as logger
import datetime
from dateutil.relativedelta import relativedelta
import math

# constants
EXACT = 0
DAYWISE = 1


# tradeList: ['id', 'date', 'type', 'coin', 'amount', 'value_*' ]
def initTradeList(data=None):
    myCoinValue = CoinValue()
    myColumns = ['id', 'date', 'type', 'coin', 'amount'] + ['value' + key for key in myCoinValue.value]
    if data:
        return pandas.DataFrame(data, columns=myColumns)
    else:
        return pandas.DataFrame(columns=myColumns)


# tradeListComplete: ['id', 'date', 'type', 'coin', 'amount', 'value_*', valueLoaded ]
def initTradeListComplete(data=None):
    myCoinValue = CoinValue()
    myColumns = ['id', 'date', 'type', 'coin', 'amount'] + ['value' + key for key in myCoinValue.value] \
                + ['valueLoaded', 'tradePartnerId', 'exchange', 'externId', 'wallet']
    if data:
        return pandas.DataFrame(data, columns=myColumns)
    else:
        return pandas.DataFrame(columns=myColumns)


def initFeeList():
    myCoinValue = CoinValue()
    return pandas.DataFrame(columns=['id', 'date', 'coin', 'fee'] + ['value' + key for key in myCoinValue.value])


# coinListComplete: [ 'coin', 'balance', 'currentValue_*']
def initCoinList(data=None):
    myCoinValue = CoinValue()
    myColumns = ['coin', 'balance'] + ['currentValue' + key for key in myCoinValue.value]
    if data:
        return pandas.DataFrame(data, columns=myColumns)
    else:
        return pandas.DataFrame(columns=myColumns)


# coinListComplete: [ 'coin', 'balance', 'initialValue_*', 'currentValue_*']
def initCoinListComplete(data=None):
    myCoinValue = CoinValue()
    myColumns = ['coin', 'balance'] + ['initialValue' + key for key in myCoinValue.value] + ['currentValue' + key for
                                                                                             key in myCoinValue.value]
    if data:
        return pandas.DataFrame(data, columns=myColumns)
    else:
        return pandas.DataFrame(columns=myColumns)

# for testing
class myDict(dict):
    def __init__(self, *args, **kwargs):
        super(myDict, self).__init__(self, *args, **kwargs)

    def __getitem__(self, key):
        return super(myDict, self).__getitem__(key)

    def __setitem__(self, key, value):
        super(myDict, self).__setitem__(key, value)

# %% CoinValue
class CoinValue():
    def __init__(self):
        self.value = {}
        try:
            for currency in settings.mySettings.displayCurrencies():
                self.value[currency] = 0
        except Exception as ex:
            logger.globalLogger.error('error initializing display currencies: ' + str(ex))
            logger.globalLogger.info('using default display currencies EUR, USD, BTC')
            self.value = {'EUR': 0, 'USD': 0, 'BTC': 0}

    # iterate values
    def __iter__(self):
        return iter(self.value)

    # return length of values
    def __len__(self):
        return len(self.value)

    # get item using []
    def __getitem__(self, key):
        return self.value[key.upper()]

    def __setitem__(self, key, value):
        self.value[key.upper()] = value

    def __lt__(self, other):
        val = 1
        for key in self.value:
            if key in other.value and self.value[key] != 0 and other.value[key] != 0:
                val *= self.value[key]/other.value[key]
        return val < 1

    def __le__(self, other):
        val = 1
        for key in self.value:
            if key in other.value:
                if other.value[key] != 0:
                    val *= self.value[key] / other.value[key]
                else:
                    if self.value[key] < 0:
                        return True
                    if self.value[key] > 0:
                        return False
        return val <= 1

    def __eq__(self, other):
        val = 1
        for key in self.value:
            if key in other.value:
                if other.value[key] != 0:
                    val *= self.value[key] / other.value[key]
                else:
                    if self.value[key] != 0:
                        return False
        return val == 1

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    # add another CoinValue to the existing one
    def add(self, other):
        for key in self.value:
            if key in other:
                self.value[key] += other.value[key]
            else:
                logger.globalLogger.error('display currency mismatch in database (add). This should not happen. Please restart application!')
        return self

    # add another CoinValue and return a new object
    def __add__(self, other):
        return self.copy().add(other)

    # substract another CoinValue from the existing one
    def sub(self, other):
        for key in self.value:
            if key in other:
                self.value[key] -= other.value[key]
            else:
                logger.globalLogger.error('display currency mismatch in database (sub). This should not happen. Please restart application!')
        return self

    # sub another CoinValue and return a new object
    def __sub__(self, other):
        return self.copy().sub(other)

    def fromDict(self, dictionary):
        for key in self.value:
            try:
                self.value[key] = dictionary[key.upper()]
            except KeyError:
                pass
            try:
                self.value[key] = dictionary[key.lower()]
            except KeyError:
                pass

        return self

    # set value of all CoinValues
    def setValue(self, value):
        for key in self.value:
            self.value[key] = value
        return self

    # devide CoinValue and return new object
    def div(self, divider):
        coinValue = CoinValue()
        if isinstance(divider, CoinValue):
            for key in self.value:
                if key in divider:
                    if divider.value[key] != 0:
                        coinValue.value[key] = self.value[key] / divider.value[key]
                    elif self.value[key] == 0:  # if both values are zero return 1
                        coinValue.value[key] = 1
                else:
                    logger.globalLogger.error(
                        'display currency mismatch in database (div). This should not happen. Please restart application!')
        else:
            if not divider == 0:
                for key in self.value:
                    coinValue.value[key] = self.value[key] / divider
                return coinValue
        return coinValue

    # multiply CoinValue and return new object
    def mult(self, multiplicator):
        coinValue = CoinValue()
        for key in self.value:
            coinValue.value[key] = self.value[key] * multiplicator
        return coinValue

    def copy(self):
        coinValue = CoinValue()
        for key in self.value:
            coinValue.value[key] = self.value[key]
        return coinValue


# %% Trade
class Trade:
    def __init__(self):
        self.tradeID = ''
        self.approxID = ''
        self.externId = ''
        self.date = None
        self.tradeType = ''
        self.coin = ''
        self.amount = None
        self.value = CoinValue()
        self.valueLoaded = False
        self.tradePartnerId = ''
        self.exchange = ''
        self.wallet = ''

    def generateID(self):
        tradeString = str(self.date) + str(self.tradeType) + str(self.externId) + str(self.coin) + str(self.amount)
        self.tradeID = hashlib.sha1(tradeString.encode()).hexdigest()[0:16]
        return self.tradeID

    def checkApproximateEquality(self, trade):
        return self.generateApproximateID() == trade.generateApproximateID()

    def generateApproximateID(self):
        myDate = converter.roundTimeMin(self.date)
        try:
            myAmount = round(self.amount, -int( math.floor(math.log10(abs(self.amount)))) + 5)
        except ValueError:
            myAmount = self.amount
        tradeString = str(str(myDate) + str(self.tradeType) + str(self.coin) + str(myAmount))
        self.approxID = hashlib.sha1(tradeString.encode()).hexdigest()
        return self.approxID

    def getApproximateID(self):
        if self.approxID:
            return self.approxID
        else:
            return self.generateApproximateID()

    def __eq__(self, other):
        return self.tradeID == other.tradeID

    def __ne__(self, other):
        return (not self.__eq__(other))

    def __hash__(self):
        return hash(self.tradeID)


    def copy(self):
        trade = Trade()
        trade.tradeID = self.tradeID
        trade.externId = self.externId
        trade.tradePartnerId = self.tradePartnerId
        trade.date = self.date
        trade.tradeType = self.tradeType
        trade.coin = self.coin
        trade.amount = self.amount
        trade.value = self.value.copy()
        trade.valueLoaded = self.valueLoaded
        trade.exchange = self.exchange
        trade.wallet = self.wallet
        return trade

    def setAmount(self, amount):
        self.value = self.value.div(self.amount).mult(amount)
        self.amount = amount

    def mergeTrade(self, trade):
        self.value.add(trade.value)
        self.amount = self.amount + trade.amount

    def toList(self):
        return [self.tradeID, self.date, self.tradeType, self.coin, self.amount] + \
               [self.getValue().value[key] for key in CoinValue().value]

    def getHeaderComplete(self):
        return ['id', 'date', 'type', 'coin', 'amount'] + ['value' + key for key in CoinValue().value] + \
               ['valueLoaded', 'tradePartnerId', 'exchange', 'externId', 'wallet']

    def toListComplete(self):
        return self.toList() + [self.valueLoaded, self.tradePartnerId, self.exchange, self.externId, self.wallet]

    def toDataFrame(self):
        return initTradeList([self.toList()])

    def updateValue(self):
        return ccapi.updateTradeValue(self)

    def getPrice(self):
        return self.value.div(self.amount)

    def getValue(self):
        return self.value

    def setValue(self, key, value):
        self.value[key] = value

    def setValueAll(self, value):
        self.value = value

    def isFiat(self):
        for fiat in settings.mySettings.fiatList():
            if fiat == self.coin:
                return True
        return False

    def toString(self):
        return str(self.toList())


# %% TradeList
class TradeList:
    def __init__(self):
        self.trades = []

    def __getitem__(self, index):
        return self.trades[index]

    def __setitem__(self, index, value):
        self.trades[index] = value

    # iterate trades
    def __iter__(self):
        return iter(self.trades)

        # return length of trades

    def __len__(self):
        return len(self.trades)

    def addTrade(self, trade):
        for mytrade in self.trades:
            if mytrade == trade:
                return False
        self.trades.append(trade)
        return True

    def addTradePair(self, trade1, trade2):
        # check if trades are already included
        trade1Index = None
        trade2Index = None
        for tradeIndex in range(len(self.trades)):
            if self.trades[tradeIndex] == trade1:
                trade1Index = tradeIndex
                if trade2Index:
                    break  # both tades found
            if self.trades[tradeIndex] == trade2:
                trade2Index = tradeIndex
                if trade1Index:
                    break  # both trades found
        # both trades found
        if trade1Index and trade2Index:
            # link found trades
            self.trades[trade1Index].tradePartnerId = self.trades[trade2Index].tradeID
            self.trades[trade2Index].tradePartnerId = self.trades[trade1Index].tradeID
            return False
        # only one trade found
        if trade1Index:
            self.trades[trade1Index].tradePartnerId = trade2.tradeID
            trade2.tradePartnerId = self.trades[trade1Index].tradeID
            self.trades.append(trade2)
        elif trade2Index:
            trade1.tradePartnerId = self.trades[trade2Index].tradeID
            self.trades[trade2Index].tradePartnerId = trade1.tradeID
            self.trades.append(trade1)
        else:  # no trade found
            trade1.tradePartnerId = trade2.tradeID
            trade2.tradePartnerId = trade1.tradeID
            self.trades.append(trade1)
            self.trades.append(trade2)
        return True

    def mergeTradeList(self, tradeList):
        addedTrades = TradeList()
        for trade in tradeList:
            if self.addTrade(trade):
                addedTrades.addTrade(trade)
        return addedTrades

    def reduceTradeList(self):
        self.trades = list(set(self.trades))

    def isEmpty(self):
        if self.trades:
            return False
        else:
            return True

    def toDataFrame(self):
        tradeList = []
        for trade in self.trades:
            tradeList.append(trade.toList())
        return initTradeList(tradeList)

    def toDataFrameComplete(self):
        trades = []
        for trade in self.trades:
            trades.append(trade.toListComplete())
        return initTradeListComplete(trades)

    def generateIDs(self):
        for trade in self.trades:
            if not trade.tradeID:
                trade.generateID()

    def reloadValues(self):
        for trade in self.trades:
            trade.updateValue()

    def setHistPrices(self, prices):
        for trade in self.trades:
            if trade.tradeID in prices:
                # check if all needed currencies are included in loaded price. Otherwise skip the trade
                missingCoin = 0
                for key in settings.mySettings.displayCurrencies():
                    if key not in prices[trade.tradeID]:
                        missingCoin = 1
                        break
                if missingCoin == 0:
                    coinPrice = CoinValue()
                    coinPrice.fromDict(prices[trade.tradeID])
                    coinVal = coinPrice.mult(trade.amount)
                    try:
                        for key in settings.mySettings.displayCurrencies():
                            trade.value[key] = coinVal[key]
                        trade.valueLoaded = True
                    except KeyError:
                        trade.valueLoaded = False

        for trade in self.trades:
            # get partner trade
            partner = self.getTradeById(trade.tradePartnerId)
            # check valid partner
            if partner:
                # check value of partner
                if partner.valueLoaded:
                    # use partner value if value update was not possible or if trade has a fiat partner
                    if (not trade.valueLoaded) or partner.isFiat():
                        trade.setValueAll(partner.getValue().mult(-1))
                        trade.valueLoaded = True
                    # if value is 0 and partner value not
                    elif trade.getValue() == CoinValue() and partner.getValue() != CoinValue():
                        trade.setValueAll(partner.getValue().mult(-1))
                    # if both values loaded and both trades are crypto use the same value
                    elif not trade.isFiat():
                        if trade.getValue() < partner.getValue():  # use smaller value (tax will be paid later)
                            partner.setValueAll(trade.getValue().mult(-1))
                        else:
                            trade.setValueAll(partner.getValue().mult(-1))

    def updateValues(self):
        # load values of all trades
        for trade in self.trades:
            if not trade.valueLoaded:
                trade.updateValue()

        for trade in self.trades:
            # get partner trade
            partner = self.getTradeById(trade.tradePartnerId)
            # check valid partner
            if partner:
                # check value of partner
                if partner.valueLoaded:
                    # use partner value if value update was not possible or if trade has a fiat partner
                    if (not trade.valueLoaded) or partner.isFiat():
                        trade.setValueAll(partner.getValue().mult(-1))
                        trade.valueLoaded = True
                    # if both values loaded and both trades are crypto use the same value
                    elif not trade.isFiat():
                        if trade.getValue() < partner.getValue():  # use smaller value (tax will be paid later)
                            partner.setValueAll(trade.getValue().mult(-1))
                        else:
                            trade.setValueAll(partner.getValue().mult(-1))


    def getTradeById(self, tradeId):
        for trade in self.trades:
            if trade.tradeID == tradeId:
                return trade
        return None

    def updateValuesAsync(self):
        t = threading.Thread(target=self.updateValues())
        t.start()
        return t

    def toCsv(self, path):
        # self.toDataFrameComplete().to_csv(path)
        with open(path, 'w') as file:
            file.write(','.join(Trade().getHeaderComplete()) + '\n')
            for trade in self:
                file.write(','.join(str(e) for e in trade.toListComplete()) + '\n')

    def clearPriceFlag(self):
        for trade in self.trades:
            trade.valueLoaded = False

    def checkApproximateEquality(self, trade):
        for t in self.trades:
            if trade.getApproximateID() == t.getApproximateID():
                return True
        return False

    def recalcIds(self):
        for trade in self.trades:
            trade.generateID()
            trade.generateApproximateID()
            if trade.tradePartnerId:
                partner = self.getTradeById(trade.tradePartnerId)
                if partner:
                    partner.tradePartnerId = trade.tradeID


# %% TradeMatcher
class TradeMatcher:
    def __init__(self, coinBalance, trades=[]):
        self.coinBalance = coinBalance
        self.trades = trades
        self.sellsBuffer = []
        self.buysBuffer = []
        self.sellsMatched = []
        self.buysMatched = []
        self.profitMatched = []
        self.profitMatchedTime = []
        self.profitMatchedBuyTime = []
        self.buysLeft = []

    def setTrades(self, trades):
        self.trades = trades

    def prepareTrades(self, timemode=DAYWISE):
        # copy trades to buffer
        self.sellsBuffer = []
        self.buysBuffer = []
        for trade in self.trades:
            # if trade.tradeType == 'trade': # todo: define tradeTypes to be used for profit calculation
            if trade.valueLoaded:  # ignore trades without loaded value
                if trade.amount < 0:
                    # copy trade to sell list
                    self.sellsBuffer.append(trade.copy())
                    # use positive sign for matching
                    self.sellsBuffer[-1].setAmount(-self.sellsBuffer[-1].amount)
                else:  # trade.amount >= 0
                    # copy trade to buy list
                    self.buysBuffer.append(trade.copy())
        if self.sellsBuffer:
            # sort trades by date
            self.sellsBuffer.sort(key=lambda x: x.date, reverse=False)
            # check timemode
            if not timemode == EXACT:  # timemode == DAYWISE or other:
                # revert all datetimes to dates
                for sell in self.sellsBuffer:
                    sell.date = sell.date.date()
            # merge sells with same date
            sellsTemp = [self.sellsBuffer[0]]
            for sell in self.sellsBuffer[1:]:
                if sellsTemp[-1].date == sell.date:
                    sellsTemp[-1].mergeTrade(sell)
                else:
                    sellsTemp.append(sell)
            self.sellsBuffer = sellsTemp
        if self.buysBuffer:
            # sort trades by date
            self.buysBuffer.sort(key=lambda x: x.date, reverse=False)
            # check timemode
            if not timemode == EXACT:  # timemode == DAYWISE or other:
                # revert all datetimes to dates
                for buy in self.buysBuffer:
                    buy.date = buy.date.date()
            # merge buys with same date
            buysTemp = [self.buysBuffer[0]]
            for buy in self.buysBuffer[1:]:
                if buysTemp[-1].date == buy.date:
                    buysTemp[-1].mergeTrade(buy)
                else:
                    buysTemp.append(buy)
            self.buysBuffer = buysTemp

    def matchTrades(self, timemode=DAYWISE):
        self.prepareTrades(timemode)
        self.sellsMatched = []
        self.buysMatched = []
        self.profitMatched = []
        self.profitMatchedTime = []
        self.profitMatchedBuyTime = []
        self.buysLeft = []
        buyIndex = 0
        sellIndex = 0
        maxIter = 10 * len(self.sellsBuffer)
        while (sellIndex < len(self.sellsBuffer) and buyIndex < len(self.buysBuffer)):
            if self.sellsBuffer[sellIndex].date < self.buysBuffer[buyIndex].date:  # sell is bevore buy
                # no buy trade can be matched -> ignore sellTrade
                logger.globalLogger.warning('no earlier buyTrade available for sellTrade: ' + self.sellsBuffer[
                    sellIndex].toString() + '; sellTrade will be ignored for profit calculation')
                sellIndex += 1
                continue
            else:  # sell is after buy
                # sell date can be matched to buy trade
                if self.sellsBuffer[sellIndex].amount == self.buysBuffer[buyIndex].amount:  # sell and buy equal
                    # save both as matched trades
                    self.sellsMatched.append(self.sellsBuffer[sellIndex])
                    self.buysMatched.append(self.buysBuffer[buyIndex])
                    # next buy trade
                    buyIndex += 1
                    # next sell trade
                    sellIndex += 1
                elif self.sellsBuffer[sellIndex].amount < self.buysBuffer[buyIndex].amount:  # sell smaller than buy
                    # copy buy trade and set amount to sell amount
                    buyMatched = self.buysBuffer[buyIndex].copy()
                    buyMatched.setAmount(self.sellsBuffer[sellIndex].amount)
                    # reduce buy trade by sell amount
                    self.buysBuffer[buyIndex].setAmount(
                        self.buysBuffer[buyIndex].amount - self.sellsBuffer[sellIndex].amount)

                    # save sell and part of buy as matched
                    self.sellsMatched.append(self.sellsBuffer[sellIndex])
                    self.buysMatched.append(buyMatched)
                    # next sell
                    sellIndex += 1
                elif self.sellsBuffer[sellIndex].amount > self.buysBuffer[buyIndex].amount:  # sell bigger than buy
                    # copy sell trade and set amount to buy amount
                    sellMatched = self.sellsBuffer[sellIndex].copy()
                    sellMatched.setAmount(self.buysBuffer[buyIndex].amount)
                    # reduce buy trade by sell amount
                    self.sellsBuffer[sellIndex].setAmount(
                        self.sellsBuffer[sellIndex].amount - self.buysBuffer[buyIndex].amount)

                    # save part of sell and buy as matched
                    self.sellsMatched.append(sellMatched)
                    self.buysMatched.append(self.buysBuffer[buyIndex])
                    # next buy
                    buyIndex += 1

                # calculate profit of last matched trades
                self.profitMatched.append(self.calculateProfit(self.buysMatched[-1], self.sellsMatched[-1]))
                self.profitMatchedTime.append(self.sellsMatched[-1].date)
                self.profitMatchedBuyTime.append(self.buysMatched[-1].date)
            maxIter -= 1
            if maxIter <= 0:
                logger.globalLogger.warning('maximum of matching iterations reached for coin ' + str(self.trades[0].coin))
                logger.globalLogger.info('current sell index: ' + str(sellIndex) + ' from ' + str(len(self.sellsBuffer)))
                logger.globalLogger.info('current buy index: ' + str(buyIndex) + ' from ' + str(len(self.buysBuffer)))
                break

        self.buysLeft = self.buysBuffer[buyIndex:]

    def calculateProfit(self, buyTrade, sellTrade):
        return sellTrade.getValue().copy().sub(buyTrade.getValue())

    def getTotalProfit(self):
        return sum(self.profitMatched, CoinValue())

    # def getTimeInvest(self, date):


    def getTimeDeltaProfit(self, fromDate, toDate, taxFreeTimeDelta=-1):
        if taxFreeTimeDelta == -1:
            return sum([profit for profit, date in zip(self.profitMatched, self.profitMatchedTime)
                        if (date >= fromDate and date <=toDate)], CoinValue())
        else:
            return sum([profit for profit, date, buydate in zip(self.profitMatched, self.profitMatchedTime,
                                                                  self.profitMatchedBuyTime)
                        if (date >= fromDate and date <= toDate and (date - relativedelta(years=taxFreeTimeDelta) <= buydate))],
                       CoinValue())

    def getInitialPrice(self):
        initialValue = CoinValue()
        initialAmount = 0
        for buy in self.buysLeft:
            initialValue.add(buy.value)
            initialAmount += buy.amount
        return initialValue.div(initialAmount)

    def toDataFrame(self, currency):
        rows = []
        for i in range(len(self.profitMatched)):
            buy = self.buysMatched[i]
            sell = self.sellsMatched[i]
            profit = self.profitMatched[i]
            rows.append([buy.date, buy.amount, buy.getPrice()[currency], buy.value[currency], \
                         sell.date, sell.amount, sell.getPrice()[currency], sell.value[currency], \
                         profit[currency]])
        return pandas.DataFrame(rows, columns=['buyDate', 'buyAmount', 'buyPrice', 'buyValue', \
                                               'sellDate', 'sellAmount', 'sellPrice', 'sellValue', \
                                               'profit'])

    def getBuyAmount(self):
        amount = 0
        for trade in self.trades:
            if trade.amount > 0:
                amount += trade.amount
        return amount

    def getSellAmount(self):
        amount = 0
        for trade in self.trades:
            if trade.amount < 0:
                amount += trade.amount
        return amount

    def getBuyAmountLeft(self):
        amount = 0
        for trade in self.buysLeft:
            amount += trade.amount
        return amount

    def getFirstBuyLeftDate(self):
        if self.buysLeft:
            return self.buysLeft[0].date
        return None

    def getBuyAmountLeftToDate(self, date):
        amount = 0
        for trade in self.buysLeft:
            try:
                if trade.date < date:
                    amount += trade.amount
            except TypeError:
                if trade.date.date() < date:
                    amount += trade.amount
        return amount

    def getBuyAmountLeftTaxFree(self, taxfreeLimit):
        return self.getBuyAmountLeftToDate(datetime.datetime.now().date() - relativedelta(years=1))


# %% Coin_Balance
class CoinBalance:
    def __init__(self):
        self.coinname = None
        self.balance = 0
        self.initialValue = CoinValue()
        self.currentPrice = CoinValue()
        self.change24h = CoinValue()
        self.trades = []
        self.tradeMatcher = TradeMatcher(self)
        self.coinIcon = None
        self.notes = ""
        self.priceChartData = []

    def __lt__(self, other):
        return self.balance < other.balance

    def __le__(self, other):
        return self.balance <= other.balance

    def __eq__(self, other):
        return self.balance == other.balance

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __comp__(self, other):
        if self.__eq__(other):
            return 0
        elif self.__lt__(other):
            return -1
        else:
            return 1

    def addTrade(self, trade):
        if self.coinname == trade.coin:
            # check if trade already exists
            for myTrade in self.trades:
                if myTrade == trade:
                    return None
            # if trade doesnt exist then add it
            self.trades.append(trade)
            # update balance (use current price)
            self.balance += trade.amount
            return self
        else:
            return None

    def removeTrade(self, trade):
        if self.coinname == trade.coin:
            # check if trade exists
            if trade in self.trades:
                self.trades.remove(trade)
            # update balance (use current price)
            self.updateBalance()
            return self
        else:
            return None

    def updateBalance(self):
        self.balance = 0
        for trade in self.trades:
            self.balance += trade.amount

    def isFiat(self):
        for fiat in settings.mySettings.fiatList():
            if fiat == self.coinname:
                return True
        return False

    def matchTrades(self):
        if self.isFiat():  # do not match fiat trades, just sum up all trades
            self.initialValue.setValue(0)
            for trade in self.trades:
                self.initialValue += trade.value
        else:  # all other trades have to be matched (sell can only be matched to an older buy)
            self.tradeMatcher.setTrades(self.trades)
            self.tradeMatcher.matchTrades()
            # check balance with buys left in tradeMatcher
            #        if self.balance != self.tradeMatcher.getBuyAmountLeft():
            # something is wrong. balance should be equal
            #            print('buys left in tradeMathcher do not fit to calculated balance for coin: ' + self.coinname)
            #            print(str(self.balance) + '; ' + str(self.tradeMatcher.getBuyAmountLeft()))
            #            pass
            # todo: create logging entry
            self.initialValue = self.tradeMatcher.getInitialPrice().mult(self.balance)

    def getInitialPrice(self):
        return self.initialValue.div(self.balance)

    def getCurrentPrice(self):
        return self.currentPrice.copy()

    def getCurrentValue(self):
        return self.currentPrice.mult(self.balance)

    def setCurrentValueAll(self, value):
        self.currentPrice = value.div(self.balance)

    def getFees(self):
        fees = []
        for trade in self.trades:
            if trade.tradeType == "fee":
                fees.append(trade)
        return fees

    def getTotalFees(self):
        fees = CoinValue()
        for trade in self.trades:
            if trade.tradeType == "fee":
                fees.add(trade.getValue())
        return fees

    def toList(self):
        return [self.coinname, self.balance] + [self.getCurrentValue().value[key] for key in self.getCurrentValue().value]

    def toListComplete(self):
        return [self.coinname, self.balance] + [self.initialValue.value[key] for key in self.initialValue.value] + \
               [self.getCurrentValue().value[key] for key in self.getCurrentValue().value]

    def getChange24h(self, key):
        if self.change24h[key]:
            return self.change24h[key]
        else:
            return 0


# %% CoinList
class CoinList:
    def __init__(self):
        self.coins = []

    # iterate coins
    def __iter__(self):
        return iter(self.coins)

        # return length of coins

    def __len__(self):
        return len(self.coins)

    # get item by name        
    def __getitem__(self, key):
        return self.getCoinByName(key)

    def hasMember(self, coinname):
        for coin in self.coins:
            if coin.coinname == coinname.upper():
                return True
        return False

    def addCoin(self, coinname):
        self.coins.append(CoinBalance())
        self.coins[-1].coinname = str(coinname)
        # self.coins[-1].coinIcon = ccapi.getIcon(coinname)
        return self.coins[-1]

    def addTrade(self, trade):
        for coin in self.coins:
            if coin.coinname == trade.coin:
                coin.addTrade(trade)
                return coin
        self.addCoin(trade.coin)
        self.coins[-1].addTrade(trade)
        return self.coins[-1]

    def removeTrade(self, trade):
        for coin in self.coins:
            if coin.coinname == trade.coin:
                coin.removeTrade(trade)
                return coin

    def addTrades(self, trades):
        for trade in trades:
            self.addTrade(trade)
        self.matchTrades()
        return self

    def deleteTrades(self, trades):
        for trade in trades:
            self.removeTrade(trade)
        # remove all empty coins
        # indexes = [index for index in range(len(self.coins)) if not self.coins[index].trades]
        # indexes.sort(reverse=True)
        # for index in indexes:
        #     self.coins.pop(index)
        self.matchTrades()
        return self

    def reloadTrades(self, trades):
        self.coins.clear()
        self.addTrades(trades)

    def tradeChanged(self, trade):
        try:
            coin = self.getCoinByName(trade.coin)
        except KeyError as ex:
            # unknowen coin; ignore trade
            # print(str(ex) + ' for coin ' + str(trade.coin))
            return None
        coin.matchTrades()
        return coin

    def matchTrades(self):
        for coin in self.coins:
            coin.matchTrades()

    def toDataFrame(self):
        coinList = []
        for coin in self.coins:
            coinList.append(coin.toList())
        return initCoinList(coinList)

    def toDataFrameComplete(self):
        coinList = []
        for coin in self.coins:
            coinList.append(coin.toListComplete())
        return initCoinListComplete(coinList)

    def isEmpty(self) -> bool:
        if self.coins:
            return False
        else:
            return True

    def getCoinByName(self, name: str) -> CoinBalance:
        for coin in self:
            if coin.coinname == name:
                return coin
        raise KeyError

    # get List of coinnames
    def getCoinNames(self):
        return [coin.coinname for coin in self.coins]

    # set prices from dict
    def setPrices(self, prices):
        for coin in self.coins:
                for key in coin.currentPrice:
                    try:
                        coin.currentPrice[key] = prices[coin.coinname][key]['PRICE']
                        coin.change24h[key] = prices[coin.coinname][key]['CHANGEPCT24HOUR']
                    except KeyError:
                        pass

    def setPriceChartData(self, priceChartData: dict):
        for coin in self.coins:
            try:
                coin.priceChartData = priceChartData[coin.coinname]
            except KeyError:
                pass


    def histPricesChanged(self):
        pass

    def histPriceUpdateFinished(self):
        self.matchTrades()

    # update current value of all coins
    def updateCurrentValues(self):
        return ccapi.updateCurrentCoinValues(self)

    def update24hChange(self):
        return ccapi.get24hChange(self)

    # update current value of all coins async
    def updateCurrentValuesAsync(self):
        t = threading.Thread(target=self.updateCurrentValues)
        t.daemon = True
        t.start()
        return t



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  4 15:15:19 2018

@author: Martin
"""

# todo: check why default section is written to file ?!


import koalafolio.gui.QLogger as logger
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtCore as qtcore
import PyQt5.QtGui as qtgui
import koalafolio.PcpCore.settings as settings

qt = qtcore.Qt
localLogger = logger.globalLogger


class QSettings(settings.Settings):
    def __init__(self, *args, **kwargs):
        super(QSettings, self).__init__(*args, **kwargs)

    def initSettings(self):
        # set default settings

        # window settings
        self['window'] = {}
        self['window']['windowsize'] = '1400x600'
        self['window']['windowpos'] = '200,200'
        self['window']['windowstate'] = '0'
        self['window']['windowTitle'] = 'KoalaFolio'
        self['window']['windowstyle'] = 'Fusion'
        self['window']['stylesheetname'] = 'defaultStyle'
        # color settings
        self['color'] = {}
        self['color']['BACKGROUND'] = '42,46,51'
        self['color']['TEXT_NORMAL'] = '255,255,255'
        self['color']['TEXT_HIGHLIGHTED'] = '42,46,51'
        self['color']['PRIMARY'] = '75,180,255'
        self['color']['SECONDARY'] = '255,105,75'
        self['color']['TERTIARY'] = '75,255,240'
        self['color']['NEGATIV'] = '255,90,75'
        self['color']['POSITIV'] = '90,255,75'
        self['color']['NEUTRAL'] = '200,200,200'
        # gui settings
        self['gui'] = {}
        self['gui']['filterUseRegex'] = 'True'
        self['gui']['portfolioFilterRow'] = '3'
        self['gui']['tradeFilterRow'] = '2'
        self['gui']['portfolioFilterDir'] = (str(qt.DescendingOrder))
        self['gui']['tradeFilterDir'] = str(qt.AscendingOrder)
        self['gui']['toolTipsEnabled'] = 'True'
        self['gui']['performanceChartIndex'] = '0'
        self['gui']['tradesEditLock'] = 'True'
        self['gui']['hideLowBalanceCoins'] = 'True'
        self['gui']['hideLowValueCoins'] = 'False'
        self['gui']['lowValueFilterLimit(reportCurrency)'] = '50'
        self['gui']['loadPriceHistoryChart'] = 'False'

        super(QSettings, self).initSettings()

    def getWindowProperties(self):
        windowProperties = {}
        try:
            size = [int(s) for s in self['window']['windowsize'].split('x')]
            pos = [int(s) for s in self['window']['windowpos'].split(',')]
        except Exception as ex:
            localLogger.error('error parsing window geometry: ' + str(ex))
            localLogger.info('using default window geometry')
            size = [1200, 800]
            pos = [200, 200]
        windowProperties['geometry'] = qtcore.QRect(pos[0], pos[1], size[0], size[1])
        try:
            state = qt.WindowState(int(self['window']['windowstate']))
        except Exception as ex:
            localLogger.error('error parsing window state: ' + str(ex))
            localLogger.info('using default window state')
            state = qt.WindowNoState
        windowProperties['state'] = state
        return windowProperties

    def setWindowProperties(self, geometry, state):
        if state == qt.WindowNoState:  # only set window size if window is in normal state
            self['window']['windowpos'] = str(geometry.x()) + ',' + str(geometry.y())
            self['window']['windowSize'] = str(geometry.width()) + 'x' + str(geometry.height())
        self['window']['windowState'] = str(int(state))

    def getColors(self):
        def convertColor(colorStr, *default):
            try:
                color = colorStr.split(',')
                r = int(color[0])
                g = int(color[1])
                b = int(color[2])
                return r, g, b  # return tuple
            except Exception as ex:
                localLogger.warning('error parsing color ' + colorStr + ': ' + str(ex))
                localLogger.info('using default color')
                return default

        colorDict = {}
        for key in self['color']:
            if key == 'default':
                continue
            if key == 'TEXT_NORMAL':
                colorDict[key.upper()] = convertColor(self['color'][key], 0, 0, 0)
                continue
            colorDict[key.upper()] = convertColor(self['color'][key], 255, 255, 255)
        return colorDict

    def getColor(self, name):
        colors = self.getColors()
        if name in colors:
            return colors[name]
        else:
            return 255, 255, 255  # return white if invalid colorname

    def setColor(self, name, *value):
        for key in self['colors']:
            if key.lower() == name.lower():
                self['colors'][key] = str(value).replace('(', '').replace(')', '')

    def getGuiSettings(self):
        gui = {}
        try:
            gui['filterUseRegex'] = self.getboolean('gui', 'filterUseRegex')
        except ValueError:
            gui['filterUseRegex'] = True
        try:
            gui['portfolioFilterRow'] = self.getint('gui', 'portfolioFilterRow')
        except ValueError:
            gui['portfolioFilterRow'] = 3
        try:
            gui['tradeFilterRow'] = self.getint('gui', 'tradeFilterRow')
        except ValueError:
            gui['tradeFilterRow'] = 2
        try:
            gui['portfolioFilterDir'] = self.getint('gui', 'portfolioFilterDir')
        except ValueError:
            gui['portfolioFilterDir'] = qt.AscendingOrder
        try:
            gui['tradeFilterDir'] = self.getint('gui', 'tradeFilterDir')
        except ValueError:
            gui['tradeFilterDir'] = qt.AscendingOrder
        try:
            gui['toolTipsEnabled'] = self.getboolean('gui', 'toolTipsEnabled')
        except ValueError:
            gui['toolTipsEnabled'] = True
        try:
            gui['performanceChartIndex'] = self.getint('gui', 'performanceChartIndex')
        except ValueError:
            gui['performanceChartIndex'] = 0
        try:
            gui['tradesEditLock'] = self.getboolean('gui', 'tradesEditLock')
        except ValueError:
            gui['tradesEditLock'] = True
        try:
            gui['hideLowBalanceCoins'] = self.getboolean('gui', 'hideLowBalanceCoins')
        except ValueError:
            gui['hideLowBalanceCoins'] = True
        try:
            gui['hideLowValueCoins'] = self.getboolean('gui', 'hideLowValueCoins')
        except ValueError:
            gui['hideLowValueCoins'] = False
        try:
            gui['lowValueFilterLimit(reportCurrency)'] = self.getfloat('gui', 'lowValueFilterLimit(reportCurrency)')
        except ValueError:
            gui['lowValueFilterLimit(reportCurrency)'] = 50
        try:
            gui['loadPriceHistoryChart'] = self.getboolean('gui', 'loadPriceHistoryChart')
        except ValueError:
            gui['loadPriceHistoryChart'] = False
        return gui

    def getGuiSetting(self, key):
        return self.getGuiSettings()[key]

    def setGuiSettings(self, gui):
        for key in gui:
            if key in self['gui']:
                self['gui'][key] = str(gui[key])
            else:
                raise KeyError('invalid key ' + key + ' for gui settings')


mySettings = QSettings()

class SettingsModelItem():
    def __init__(self, key, parentkey, parent):
        self.key = key
        self.parentKey = parentkey
        self.parent = parent

    def isEqual(self, other):
        return self.key == other.key and self.parentKey == other.parentKey


# settings model
class SettingsModel(qtcore.QAbstractItemModel):
    displayCurrenciesChanged = qtcore.pyqtSignal([list])

    def __init__(self, settings, *args, **kwargs):
        super(SettingsModel, self).__init__(*args, **kwargs)
        self.settings = settings
        self.sections = list(self.settings)[1:]
        self.items = []
        self.header = ['setting', 'value']

    def headerData(self, section, orientation, role):
        if role == qt.DisplayRole:
            return self.header[section]
        return qtcore.QVariant()

    def parent(self, index):
        return self.getParentIndexFromIndex(index)

    def rowCount(self, parent):
        if not parent.isValid():
            return len(self.sections)
        else:
            # check if it is a parent or a child item
            if self.getParentIndexFromIndex(parent).isValid():  # only a child item has a valid parent
                return 0  # return no childs
            else:  # parent item, return number of childs
                return len(self.settings[self.getItemFromIndex(parent).key])

    def columnCount(self, parent):
        return 2

    def index(self, row, column, parent):
        if parent.isValid():
            parentkey = self.getItemFromIndex(parent).key
            subsettings = self.settings[parentkey]
            key = list(subsettings)[row]
        else:
            parentkey = ""
            key = self.sections[row]
        newItem = SettingsModelItem(key, parentkey, parent)
        itemId = -1
        itemFound = False
        for item in self.items:
            itemId += 1
            if item.isEqual(newItem):
                itemFound = True
                break
        if not itemFound:
            itemId = len(self.items)
            self.items.append(newItem)
        index = self.createIndex(row, column, itemId)
        return index

    def data(self, index, role):
        if role == qt.DisplayRole:
            parentIndex = self.getParentIndexFromIndex(index)
            item = self.getItemFromIndex(index)
            if parentIndex.isValid():
                parent = self.getItemFromIndex(parentIndex)
                if index.column() == 0:
                    return item.key
                if index.column() == 1:
                    return self.settings[parent.key][item.key]
            else:
                return self.sections[index.row()]
        return qtcore.QVariant()

    def setData(self, index, value, role):
        if role == qt.EditRole:
            try:
                item = self.getItemFromIndex(index)
                self.settings[item.parentKey][item.key] = value
                self.dataChanged.emit(index, index)
                self.itemChanged(item)
                return True
            except:
                return False
        return False

    def itemChanged(self, item):
        if item.parentKey.lower() == 'currency' and item.key.lower() == 'defaultdisplaycurrencies':
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())

    def resetDefault(self):
        oldCur = self.settings.displayCurrencies()
        self.settings.resetDefault()
        if oldCur != self.settings.displayCurrencies():
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())

    def restoreSettings(self):
        oldCur = self.settings.displayCurrencies()
        self.settings.readSettings()
        if oldCur != self.settings.displayCurrencies():
            self.displayCurrenciesChanged.emit(self.settings.displayCurrencies())

    def flags(self, index):
        column = index.column()
        parent = self.getParentIndexFromIndex(index)
        if parent.isValid():
            if column == 0:
                return qt.ItemIsSelectable | qt.ItemIsEnabled | qt.ItemNeverHasChildren
            if column == 1:
                return qt.ItemIsSelectable | qt.ItemIsEditable | qt.ItemIsEnabled | qt.ItemNeverHasChildren
        else:
            return qt.ItemIsSelectable | qt.ItemIsEnabled

    def getParentIndexFromIndex(self, index):
        return self.items[index.internalId()].parent

    def getItemFromIndex(self, index):
        return self.items[index.internalId()]

    def getParentFromIndex(self, index):
        parentIndex = self.getParentIndexFromIndex(index)
        if parentIndex.isValid():
            return self.getItemFromIndex(parentIndex)
        else:
            return None


class SettingsDelegate(qtwidgets.QStyledItemDelegate):
    def __init__(self, *args, **kwargs):
        super(SettingsDelegate, self).__init__(*args, **kwargs)

    def createEditor(self, parent, option, index):
        if int(index.flags()) & qt.ItemIsEditable:
            return qtwidgets.QLineEdit(parent)
        return 0

    def setEditorData(self, editor, index):
        editor.setText(index.data())

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.text(), qt.EditRole)

    def sizeHint(self, option, index):
        return qtcore.QSize()


class SettingsTreeView(qtwidgets.QTreeView):
    def __init__(self, *args, **kwargs):
        super(SettingsTreeView, self).__init__(*args, **kwargs)

        self.setItemDelegateForColumn(1, SettingsDelegate())



# flake8: noqa

import json
import os
from typing import Any, Dict, List, Optional, Tuple
from pathlib import Path
import requests

from rotkehlchen.chain.ethereum.contracts import EthereumContract
from rotkehlchen.serialization.deserialize import deserialize_ethereum_address

MAX_BLOCKTIME_CACHE = 250  # 55 mins with 13 secs avg block time
ZERO_ADDRESS = deserialize_ethereum_address('0x0000000000000000000000000000000000000000')
AAVE_ETH_RESERVE_ADDRESS = deserialize_ethereum_address('0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE')

def _get_latest_data(data_directory: Path) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Gets the latest assets either locally or from the remote

    Checks the remote (github) and if there is a newer file there it pulls it,
    saves it and its md5 hash locally and returns the new assets.

    If there is no new file (same hash) or if there is any problem contacting the remote
    then the builtin assets file is used.
    """
    our_downloaded_contracts = data_directory / 'assets' / 'eth_contracts.json'
    our_downloaded_abi_entries = data_directory / 'assets' / 'eth_abi.json'
    try:
        # we need to download and save the contracts from github
        response = requests.get(
            'https://raw.githubusercontent.com/rotki/rotki/develop/rotkehlchen/data/eth_contracts.json')  # noqa: E501
        remote_contract_data = response.text
        # we need to download and save the abi from github
        response2 = requests.get(
            'https://raw.githubusercontent.com/rotki/rotki/develop/rotkehlchen/data/eth_abi.json')
        remote_abi_data = response2.text

        # Make sure directory exists
        (data_directory / 'assets').mkdir(parents=True, exist_ok=True)
        # Write the files
        with open(our_downloaded_contracts, 'w') as f:
            f.write(remote_contract_data)
        with open(our_downloaded_abi_entries, 'w') as f:
            f.write(remote_abi_data)

        return json.loads(remote_contract_data), json.loads(remote_abi_data)

        # else, continue without new data
    except (requests.exceptions.ConnectionError, KeyError, json.decoder.JSONDecodeError):
        pass

    if our_downloaded_contracts.is_file() and our_downloaded_abi_entries.is_file():
        with open(our_downloaded_contracts, 'r') as f:
            remote_contract_data = f.read()
        with open(our_downloaded_abi_entries, 'r') as f:
            remote_abi_data = f.read()
        return json.loads(remote_contract_data), json.loads(remote_abi_data)

    # no local data and no remote data available
    return None, None


def _attempt_initialization(
        data_directory: Optional[Path]
) -> Tuple[Dict[str, Any], Dict[str, Any], bool]:
    """Reads the ethereum data from the remote

    1. If it's the very first time data is initialized (and data directory is not given)
    then just get assets from the builtin file
    2. If data directory is still not given but we have some saved assets return them directly
    3. If data directory is given then we can finally do the comparison of local
    saved and builtin file with the remote and return the most recent assets.

    Returns a tuple of the most recent data mapping it can get and a boolean denoting
    if the remote check happened or not.
    """
    if data_directory:
        # we got the data directory so we can finally do the remote check
        contracts, abi_entries = _get_latest_data(data_directory)
        return contracts, abi_entries, True
    return None, None, False

class EthereumConstants():
    __instance = None
    contracts: Dict[str, Dict[str, Any]] = {}
    abi_entries: Dict[str, List[Dict[str, Any]]] = {}

    def __new__(
            cls,
            data_directory: Path = None,
    ) -> 'EthereumConstants':
        if EthereumConstants.__instance is not None:
            if EthereumConstants.__instance.remote_check_happened:  # type: ignore
                return EthereumConstants.__instance

            # else we still have not performed the remote check
            contracts, abi_entries, check_happened = _attempt_initialization(
                data_directory=data_directory
            )
        else:
            # first initialization
            contracts, abi_entries, check_happened = _attempt_initialization(
                data_directory=data_directory
            )
            EthereumConstants.__instance = object.__new__(cls)

        EthereumConstants.__instance.contracts = contracts
        EthereumConstants.__instance.abi_entries = abi_entries
        EthereumConstants.__instance.remote_check_happened = check_happened

        return EthereumConstants.__instance

    @staticmethod
    def get() -> Dict[str, Dict[str, Any]]:
        return EthereumConstants().contracts

    @staticmethod
    def contract_or_none(name: str) -> Optional[EthereumContract]:
        """Gets details of an ethereum contract from the contracts json file

        Returns None if missing
        """
        contract = EthereumConstants().contracts.get(name, None)
        if contract is None:
            return None

        return EthereumContract(
            address=contract['address'],
            abi=contract['abi'],
            deployed_block=contract['deployed_block'],
        )

    @staticmethod
    def contract(name: str) -> EthereumContract:
        """Gets details of an ethereum contract from the contracts json file

        Missing contract is an error
        """
        contract = EthereumConstants().contract_or_none(name)
        assert contract, f'No contract data for {name} found'
        return contract

    @staticmethod
    def abi_or_none(name: str) -> Optional[List[Dict[str, Any]]]:
        """Gets abi of an ethereum contract from the abi json file

        Returns None if missing
        """
        return EthereumConstants().abi_entries.get(name, None)

    @staticmethod
    def abi(name: str) -> List[Dict[str, Any]]:
        abi = EthereumConstants().abi_or_none(name)
        assert abi, f'No abi for {name} found'
        return abi

# Latest contract addresses are in the makerdao changelog. These values are taken from here:
# https://changelog.makerdao.com/releases/mainnet/1.1.3/contracts.json


MAKERDAO_DAI_JOIN = lambda: EthereumConstants().contract('MAKERDAO_DAI_JOIN')
MAKERDAO_CDP_MANAGER = lambda: EthereumConstants().contract('MAKERDAO_CDP_MANAGER')
MAKERDAO_GET_CDPS = lambda: EthereumConstants().contract('MAKERDAO_GET_CDPS')
MAKERDAO_PROXY_REGISTRY = lambda: EthereumConstants().contract('MAKERDAO_PROXY_REGISTRY')
MAKERDAO_SPOT = lambda: EthereumConstants().contract('MAKERDAO_SPOT')
MAKERDAO_POT = lambda: EthereumConstants().contract('MAKERDAO_POT')
MAKERDAO_VAT = lambda: EthereumConstants().contract('MAKERDAO_VAT')
MAKERDAO_ETH_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_ETH_A_JOIN')
MAKERDAO_ETH_B_JOIN = lambda: EthereumConstants().contract('MAKERDAO_ETH_B_JOIN')
MAKERDAO_BAT_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_BAT_A_JOIN')
MAKERDAO_USDC_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_USDC_A_JOIN')
MAKERDAO_USDC_B_JOIN = lambda: EthereumConstants().contract('MAKERDAO_USDC_B_JOIN')
MAKERDAO_USDT_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_USDT_A_JOIN')
MAKERDAO_WBTC_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_WBTC_A_JOIN')
MAKERDAO_KNC_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_KNC_A_JOIN')
MAKERDAO_MANA_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_MANA_A_JOIN')
MAKERDAO_TUSD_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_TUSD_A_JOIN')
MAKERDAO_ZRX_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_ZRX_A_JOIN')
MAKERDAO_PAXUSD_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_PAXUSD_A_JOIN')
MAKERDAO_COMP_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_COMP_A_JOIN')
MAKERDAO_LRC_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_LRC_A_JOIN')
MAKERDAO_LINK_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_LINK_A_JOIN')
MAKERDAO_BAL_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_BAL_A_JOIN')
MAKERDAO_YFI_A_JOIN = lambda: EthereumConstants().contract('MAKERDAO_YFI_A_JOIN')

MAKERDAO_CAT = lambda: EthereumConstants().contract('MAKERDAO_CAT')
MAKERDAO_JUG = lambda: EthereumConstants().contract('MAKERDAO_JUG')

YEARN_YCRV_VAULT = lambda: EthereumConstants().contract('YEARN_YCRV_VAULT')
YEARN_3CRV_VAULT = lambda: EthereumConstants().contract('YEARN_3CRV_VAULT')
YEARN_DAI_VAULT = lambda: EthereumConstants().contract('YEARN_DAI_VAULT')
YEARN_WETH_VAULT = lambda: EthereumConstants().contract('YEARN_WETH_VAULT')
YEARN_YFI_VAULT = lambda: EthereumConstants().contract('YEARN_YFI_VAULT')
YEARN_ALINK_VAULT = lambda: EthereumConstants().contract('YEARN_ALINK_VAULT')
YEARN_USDT_VAULT = lambda: EthereumConstants().contract('YEARN_USDT_VAULT')
YEARN_USDC_VAULT = lambda: EthereumConstants().contract('YEARN_USDC_VAULT')
YEARN_TUSD_VAULT = lambda: EthereumConstants().contract('YEARN_TUSD_VAULT')
YEARN_GUSD_VAULT = lambda: EthereumConstants().contract('YEARN_GUSD_VAULT')
YEARN_BCURVE_VAULT = lambda: EthereumConstants().contract('YEARN_BCURVE_VAULT')
YEARN_SRENCURVE_VAULT = lambda: EthereumConstants().contract('YEARN_SRENCURVE_VAULT')

ETH_SCAN = lambda: EthereumConstants().contract('ETH_SCAN')
ETH_MULTICALL = lambda: EthereumConstants().contract('ETH_MULTICALL')


AAVE_LENDING_POOL = lambda: EthereumConstants().contract('AAVE_LENDING_POOL')

ATOKEN_ABI = lambda: EthereumConstants.abi('ATOKEN')
ZERION_ABI = lambda: EthereumConstants.abi('ZERION_ADAPTER')
CTOKEN_ABI = lambda: EthereumConstants.abi('CTOKEN')
ERC20TOKEN_ABI = lambda: EthereumConstants.abi('ERC20_TOKEN')
FARM_ASSET_ABI = lambda: EthereumConstants.abi('FARM_ASSET')

YEARN_VAULTS_PREFIX = 'yearn_vaults_events'



from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional

from gevent.lock import Semaphore

from rotkehlchen.accounting.structures import Balance
from rotkehlchen.assets.asset import Asset, EthereumToken
from rotkehlchen.chain.ethereum.structures import YearnVault, YearnVaultEvent
from rotkehlchen.chain.ethereum.utils import token_normalized_value
from rotkehlchen.constants.ethereum import (
    ERC20TOKEN_ABI,
    MAX_BLOCKTIME_CACHE,
    YEARN_ALINK_VAULT,
    YEARN_BCURVE_VAULT,
    YEARN_DAI_VAULT,
    YEARN_GUSD_VAULT,
    YEARN_SRENCURVE_VAULT,
    YEARN_TUSD_VAULT,
    YEARN_USDC_VAULT,
    YEARN_USDT_VAULT,
    YEARN_VAULTS_PREFIX,
    YEARN_WETH_VAULT,
    YEARN_YCRV_VAULT,
    YEARN_YFI_VAULT,
    ZERO_ADDRESS,
)
from rotkehlchen.constants.misc import ZERO
from rotkehlchen.errors import UnknownAsset
from rotkehlchen.fval import FVal
from rotkehlchen.history.price import query_usd_price_zero_if_error
from rotkehlchen.inquirer import SPECIAL_SYMBOLS, Inquirer
from rotkehlchen.typing import ChecksumEthAddress, Price, Timestamp
from rotkehlchen.user_messages import MessagesAggregator
from rotkehlchen.utils.interfaces import EthereumModule
from rotkehlchen.utils.misc import address_to_bytes32, hexstr_to_int, ts_now

if TYPE_CHECKING:
    from rotkehlchen.chain.ethereum.manager import EthereumManager
    from rotkehlchen.chain.ethereum.zerion import GIVEN_DEFI_BALANCES, DefiProtocolBalances
    from rotkehlchen.db.dbhandler import DBHandler

BLOCKS_PER_YEAR = 2425846


YEARN_VAULTS = {
    'yyDAI+yUSDC+yUSDT+yTUSD': YearnVault(
        name='YCRV Vault',
        contract=YEARN_YCRV_VAULT,
        underlying_token=EthereumToken('yDAI+yUSDC+yUSDT+yTUSD'),
        token=EthereumToken('yyDAI+yUSDC+yUSDT+yTUSD'),
    ),
    'yDAI': YearnVault(
        name='YDAI Vault',
        contract=YEARN_DAI_VAULT,
        underlying_token=EthereumToken('DAI'),
        token=EthereumToken('yDAI'),
    ),
    'yWETH': YearnVault(
        name='YWETH Vault',
        contract=YEARN_WETH_VAULT,
        underlying_token=EthereumToken('WETH'),
        token=EthereumToken('yWETH'),
    ),
    'yYFI': YearnVault(
        name='YYFI Vault',
        contract=YEARN_YFI_VAULT,
        underlying_token=EthereumToken('YFI'),
        token=EthereumToken('yYFI'),
    ),
    'yaLINK': YearnVault(
        name='YALINK Vault',
        contract=YEARN_ALINK_VAULT,
        underlying_token=EthereumToken('aLINK'),
        token=EthereumToken('yaLINK'),
    ),
    'yUSDT': YearnVault(
        name='YUSDT Vault',
        contract=YEARN_USDT_VAULT,
        underlying_token=EthereumToken('USDT'),
        token=EthereumToken('yUSDT'),
    ),
    'yUSDC': YearnVault(
        name='YUSDC Vault',
        contract=YEARN_USDC_VAULT,
        underlying_token=EthereumToken('USDC'),
        token=EthereumToken('yUSDC'),
    ),
    'yTUSD': YearnVault(
        name='YTUSD Vault',
        contract=YEARN_TUSD_VAULT,
        underlying_token=EthereumToken('TUSD'),
        token=EthereumToken('yTUSD'),
    ),
    'yGUSD': YearnVault(
        name='GUSD Vault',
        contract=YEARN_GUSD_VAULT,
        underlying_token=EthereumToken('GUSD'),
        token=EthereumToken('yGUSD'),
    ),
    'yyDAI+yUSDC+yUSDT+yBUSD': YearnVault(
        name='YBCURVE Vault',
        contract=YEARN_BCURVE_VAULT,
        underlying_token=EthereumToken('yDAI+yUSDC+yUSDT+yBUSD'),
        token=EthereumToken('yyDAI+yUSDC+yUSDT+yBUSD'),
    ),
    'ycrvRenWSBTC': YearnVault(
        name='YSRENCURVE Vault',
        contract=YEARN_SRENCURVE_VAULT,
        underlying_token=EthereumToken('crvRenWSBTC'),
        token=EthereumToken('ycrvRenWSBTC'),
    ),
}


class YearnVaultHistory(NamedTuple):
    events: List[YearnVaultEvent]
    profit_loss: Balance


class YearnVaultBalance(NamedTuple):
    underlying_token: Asset
    vault_token: Asset
    underlying_value: Balance
    vault_value: Balance
    roi: FVal

    def serialize(self) -> Dict[str, Any]:
        result = self._asdict()  # pylint: disable=no-member
        result['roi'] = self.roi.to_percentage(precision=2)
        return result


def get_usd_price_zero_if_error(
        asset: Asset,
        time: Timestamp,
        location: str,
        msg_aggregator: MessagesAggregator,
) -> Price:
    """A special version of query_usd_price_zero_if_error using current price instead
    of historical token price for some assets.

    Since these assets are not supported by our price oracles we derive current
    price from the chain but without an archive node can't query old prices.

    TODO: MAke an issue about this
    This can be solved when we have an archive node.
    """
    if asset.identifier in SPECIAL_SYMBOLS:
        return Inquirer().find_usd_price(asset)

    return query_usd_price_zero_if_error(
        asset=asset,
        time=time,
        location=location,
        msg_aggregator=msg_aggregator,
    )


class YearnVaults(EthereumModule):

    def __init__(
            self,
            ethereum_manager: 'EthereumManager',
            database: 'DBHandler',
            msg_aggregator: MessagesAggregator,
    ) -> None:
        self.ethereum = ethereum_manager
        self.database = database
        self.msg_aggregator = msg_aggregator
        self.history_lock = Semaphore()

    def _calculate_vault_roi(self, vault: YearnVault) -> FVal:
        """
        getPricePerFullShare A @ block X
        getPricePerFullShare B @ block Y

        (A-B / X-Y) * blocksPerYear (2425846)

        So the numbers you see displayed on http://yearn.finance/vaults
        are ROI since launch of contract. All vaults start with pricePerFullShare = 1e18
        """
        now_block_number = self.ethereum.get_latest_block_number()
        price_per_full_share = self.ethereum.call_contract(
            contract_address=vault.contract.address,
            abi=YEARN_DAI_VAULT.abi,  # Any vault ABI will do
            method_name='getPricePerFullShare',
        )
        nominator = price_per_full_share - (10**18)
        denonimator = now_block_number - vault.contract.deployed_block
        return FVal(nominator) / FVal(denonimator) * BLOCKS_PER_YEAR / 10**18

    def _get_single_addr_balance(
            self,
            defi_balances: List['DefiProtocolBalances'],
            roi_cache: Dict[str, FVal],
    ) -> Dict[str, YearnVaultBalance]:
        result = {}
        for balance in defi_balances:
            if balance.protocol.name == 'yearn.finance • Vaults':
                underlying_symbol = balance.underlying_balances[0].token_symbol
                vault_symbol = balance.base_balance.token_symbol
                vault = YEARN_VAULTS.get(vault_symbol, None)
                if vault is None:
                    self.msg_aggregator.add_warning(
                        f'Found balance for unsupported yearn vault {vault_symbol}',
                    )
                    continue

                try:
                    underlying_asset = Asset(underlying_symbol)
                    vault_asset = Asset(vault_symbol)
                except UnknownAsset as e:
                    self.msg_aggregator.add_warning(
                        f'Found unknown asset {e.asset_name} for yearn vault entry',
                    )
                    continue

                roi = roi_cache.get(underlying_symbol, None)
                if roi is None:
                    roi = self._calculate_vault_roi(vault)
                    roi_cache[underlying_symbol] = roi

                result[vault.name] = YearnVaultBalance(
                    underlying_token=underlying_asset,
                    vault_token=vault_asset,
                    underlying_value=balance.underlying_balances[0].balance,
                    vault_value=balance.base_balance.balance,
                    roi=roi,
                )

        return result

    def get_balances(
            self,
            given_defi_balances: 'GIVEN_DEFI_BALANCES',
    ) -> Dict[ChecksumEthAddress, Dict[str, YearnVaultBalance]]:
        if isinstance(given_defi_balances, dict):
            defi_balances = given_defi_balances
        else:
            defi_balances = given_defi_balances()

        roi_cache: Dict[str, FVal] = {}
        result = {}
        for address, balances in defi_balances.items():
            vault_balances = self._get_single_addr_balance(balances, roi_cache)
            if len(vault_balances) != 0:
                result[address] = vault_balances

        return result

    def _get_vault_deposit_events(
            self,
            vault: YearnVault,
            address: ChecksumEthAddress,
            from_block: int,
            to_block: int,
    ) -> List[YearnVaultEvent]:
        """Get all deposit events of the underlying token to the vault"""
        events: List[YearnVaultEvent] = []
        argument_filters = {'from': address, 'to': vault.contract.address}
        deposit_events = self.ethereum.get_logs(
            contract_address=vault.underlying_token.ethereum_address,
            abi=ERC20TOKEN_ABI,
            event_name='Transfer',
            argument_filters=argument_filters,
            from_block=from_block,
            to_block=to_block,
        )
        for deposit_event in deposit_events:
            timestamp = self.ethereum.get_event_timestamp(deposit_event)
            deposit_amount = token_normalized_value(
                token_amount=hexstr_to_int(deposit_event['data']),
                token=vault.underlying_token,
            )
            tx_hash = deposit_event['transactionHash']
            tx_receipt = self.ethereum.get_transaction_receipt(tx_hash)
            deposit_index = deposit_event['logIndex']
            mint_amount = None
            for log in tx_receipt['logs']:
                found_event = (
                    log['topics'][0] == '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef' and  # noqa: E501
                    log['topics'][1] == address_to_bytes32(ZERO_ADDRESS) and
                    log['topics'][2] == address_to_bytes32(address)
                )
                if found_event:
                    # found the mint log
                    mint_amount = token_normalized_value(
                        token_amount=hexstr_to_int(log['data']),
                        token=vault.token,
                    )

            if mint_amount is None:
                self.msg_aggregator.add_error(
                    f'Ignoring yearn deposit event with tx_hash {tx_hash} and log index '
                    f'{deposit_index} due to inability to find corresponding mint event',
                )
                continue

            deposit_usd_price = get_usd_price_zero_if_error(
                asset=vault.underlying_token,
                time=timestamp,
                location='yearn vault deposit',
                msg_aggregator=self.msg_aggregator,
            )
            mint_usd_price = get_usd_price_zero_if_error(
                asset=vault.token,
                time=timestamp,
                location='yearn vault deposit',
                msg_aggregator=self.msg_aggregator,
            )
            events.append(YearnVaultEvent(
                event_type='deposit',
                block_number=deposit_event['blockNumber'],
                timestamp=timestamp,
                from_asset=vault.underlying_token,
                from_value=Balance(
                    amount=deposit_amount,
                    usd_value=deposit_amount * deposit_usd_price,
                ),
                to_asset=vault.token,
                to_value=Balance(
                    amount=mint_amount,
                    usd_value=mint_amount * mint_usd_price,
                ),
                realized_pnl=None,
                tx_hash=tx_hash,
                log_index=deposit_index,
            ))

        return events

    def _get_vault_withdraw_events(
            self,
            vault: YearnVault,
            address: ChecksumEthAddress,
            from_block: int,
            to_block: int,
    ) -> List[YearnVaultEvent]:
        """Get all withdraw events of the underlying token to the vault"""
        events: List[YearnVaultEvent] = []
        argument_filters = {'from': vault.contract.address, 'to': address}
        withdraw_events = self.ethereum.get_logs(
            contract_address=vault.underlying_token.ethereum_address,
            abi=ERC20TOKEN_ABI,
            event_name='Transfer',
            argument_filters=argument_filters,
            from_block=from_block,
            to_block=to_block,
        )
        for withdraw_event in withdraw_events:
            timestamp = self.ethereum.get_event_timestamp(withdraw_event)
            withdraw_amount = token_normalized_value(
                token_amount=hexstr_to_int(withdraw_event['data']),
                token=vault.token,
            )
            tx_hash = withdraw_event['transactionHash']
            tx_receipt = self.ethereum.get_transaction_receipt(tx_hash)
            withdraw_index = withdraw_event['logIndex']
            burn_amount = None
            for log in tx_receipt['logs']:
                found_event = (
                    log['topics'][0] == '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef' and  # noqa: E501
                    log['topics'][1] == address_to_bytes32(address) and
                    log['topics'][2] == address_to_bytes32(ZERO_ADDRESS)
                )
                if found_event:
                    # found the burn log
                    burn_amount = token_normalized_value(
                        token_amount=hexstr_to_int(log['data']),
                        token=vault.token,
                    )

            if burn_amount is None:
                self.msg_aggregator.add_error(
                    f'Ignoring yearn withdraw event with tx_hash {tx_hash} and log index '
                    f'{withdraw_index} due to inability to find corresponding burn event',
                )
                continue

            withdraw_usd_price = get_usd_price_zero_if_error(
                asset=vault.underlying_token,
                time=timestamp,
                location='yearn vault withdraw',
                msg_aggregator=self.msg_aggregator,
            )
            burn_usd_price = get_usd_price_zero_if_error(
                asset=vault.token,
                time=timestamp,
                location='yearn vault withdraw',
                msg_aggregator=self.msg_aggregator,
            )
            events.append(YearnVaultEvent(
                event_type='withdraw',
                block_number=withdraw_event['blockNumber'],
                timestamp=timestamp,
                from_asset=vault.token,
                from_value=Balance(
                    amount=burn_amount,
                    usd_value=burn_amount * burn_usd_price,
                ),
                to_asset=vault.underlying_token,
                to_value=Balance(
                    amount=withdraw_amount,
                    usd_value=withdraw_amount * withdraw_usd_price,
                ),
                realized_pnl=None,
                tx_hash=tx_hash,
                log_index=withdraw_index,
            ))

        return events

    def _process_vault_events(self, events: List[YearnVaultEvent]) -> Balance:
        """Process the events for a single vault and returns total profit/loss after all events"""
        total = Balance()
        profit_so_far = Balance()
        for event in events:
            if event.event_type == 'deposit':
                total -= event.from_value
            else:  # withdraws
                profit_amount = total.amount + event.to_value.amount - profit_so_far.amount
                profit: Optional[Balance]
                if profit_amount >= 0:
                    usd_price = get_usd_price_zero_if_error(
                        asset=event.to_asset,
                        time=event.timestamp,
                        location='yearn vault event processing',
                        msg_aggregator=self.msg_aggregator,
                    )
                    profit = Balance(profit_amount, profit_amount * usd_price)
                    profit_so_far += profit
                else:
                    profit = None

                event.realized_pnl = profit
                total += event.to_value

        return total

    def get_vault_history(
            self,
            defi_balances: List['DefiProtocolBalances'],
            vault: YearnVault,
            address: ChecksumEthAddress,
            from_block: int,
            to_block: int,
    ) -> Optional[YearnVaultHistory]:
        from_block = max(from_block, vault.contract.deployed_block)
        last_query = self.database.get_used_query_range(
            name=f'{YEARN_VAULTS_PREFIX}_{vault.name.replace(" ", "_")}_{address}',
        )
        skip_query = last_query and to_block - last_query[1] < MAX_BLOCKTIME_CACHE

        events = self.database.get_yearn_vaults_events(address=address, vault=vault)
        if not skip_query:
            query_from_block = last_query[1] + 1 if last_query else from_block
            new_events = self._get_vault_deposit_events(vault, address, query_from_block, to_block)
            if len(events) == 0 and len(new_events) == 0:
                # After all events have been queried then also update the query range.
                # Even if no events are found for an address we need to remember the range
                self.database.update_used_block_query_range(
                    name=f'{YEARN_VAULTS_PREFIX}_{vault.name.replace(" ", "_")}_{address}',
                    from_block=from_block,
                    to_block=to_block,
                )
                return None

            new_events.extend(
                self._get_vault_withdraw_events(vault, address, query_from_block, to_block),
            )
            # Now update the DB with the new events
            self.database.add_yearn_vaults_events(address, new_events)
            events.extend(new_events)

        # After all events have been queried then also update the query range.
        # Even if no events are found for an address we need to remember the range
        self.database.update_used_block_query_range(
            name=f'{YEARN_VAULTS_PREFIX}_{vault.name.replace(" ", "_")}_{address}',
            from_block=from_block,
            to_block=to_block,
        )
        if len(events) == 0:
            return None

        events.sort(key=lambda x: x.timestamp)
        total_pnl = self._process_vault_events(events)

        current_balance = None
        for balance in defi_balances:
            found_balance = (
                balance.protocol.name == 'yearn.finance • Vaults' and
                balance.base_balance.token_symbol == vault.token.symbol
            )
            if found_balance:
                current_balance = balance.underlying_balances[0].balance
                total_pnl += current_balance
                break

        # Due to the way we calculate usd prices for vaults we need to get the current
        # usd price of the actual pnl amount at this point
        if total_pnl.amount != ZERO:
            usd_price = get_usd_price_zero_if_error(
                asset=vault.underlying_token,
                time=ts_now(),
                location='yearn vault history',
                msg_aggregator=self.msg_aggregator,
            )
            total_pnl.usd_value = usd_price * total_pnl.amount

        return YearnVaultHistory(events=events, profit_loss=total_pnl)

    def get_history(
            self,
            given_defi_balances: 'GIVEN_DEFI_BALANCES',
            addresses: List[ChecksumEthAddress],
            reset_db_data: bool,
            from_timestamp: Timestamp,  # pylint: disable=unused-argument
            to_timestamp: Timestamp,  # pylint: disable=unused-argument
    ) -> Dict[ChecksumEthAddress, Dict[str, YearnVaultHistory]]:
        with self.history_lock:

            if reset_db_data is True:
                self.database.delete_yearn_vaults_data()

            if isinstance(given_defi_balances, dict):
                defi_balances = given_defi_balances
            else:
                defi_balances = given_defi_balances()

            from_block = self.ethereum.etherscan.get_blocknumber_by_time(from_timestamp)
            to_block = self.ethereum.etherscan.get_blocknumber_by_time(to_timestamp)
            history: Dict[ChecksumEthAddress, Dict[str, YearnVaultHistory]] = {}

            for address in addresses:
                history[address] = {}
                for _, vault in YEARN_VAULTS.items():
                    vault_history = self.get_vault_history(
                        defi_balances=defi_balances.get(address, []),
                        vault=vault,
                        address=address,
                        from_block=from_block,
                        to_block=to_block,
                    )
                    if vault_history:
                        history[address][vault.name] = vault_history

                if len(history[address]) == 0:
                    del history[address]

        return history

    # -- Methods following the EthereumModule interface -- #
    def on_startup(self) -> None:
        pass

    def on_account_addition(self, address: ChecksumEthAddress) -> None:
        pass

    def on_account_removal(self, address: ChecksumEthAddress) -> None:
        pass
